"""
Main class of the interact module, which derives from :class:`Deal` but keeps track
of changes to its state.
"""

__all__ = ["InteractiveDeal"]

from endplay.types import Card, Deal, Player, Denom

class InteractiveDeal(Deal):
	"""
	A Deal which keeps track of changes made to it allowing actions to be undone.
	This is the basis of the various frontends to the interact package.
	"""
	def __init__(self, pbn: str = None, first: Player = Player.north, trump: Denom = Denom.nt):
		self._history = []
		super().__init__(pbn, first, trump)

	@Deal.trump.setter
	def trump(self, denom: Denom) -> None:
		self._history.append(("trump", self.trump))
		self._data.trump = denom

	def _undo_trump(self, denom):
		self._data.trump = denom

	@Deal.first.setter
	def first(self, player: Player) -> None:
		self._history.append(("first", self.first))
		self._data.first = player

	def _undo_first(self, player):
		self._data.first = player

	def play(self, card, fromHand: bool = True) -> None:
		if len(self.curtrick) == 3:
			toundo = ("play", (fromHand, self.first, self.curtrick, card))
		else:
			toundo = ("play", (fromHand,))
		super().play(card)
		self._history.append(toundo)
		
	def _undo_play(self, arg):
		if len(arg) == 1:
			self.unplay(arg)
		else:
			fromHand, first, trick, last_card = arg
			self._data.first = first
			for card in trick:
				super().play(card, False)
			if fromHand:
				super().__getitem__(self.first.rho).add(last_card)

	def unplay(self, toHand: bool = True) -> Card:
		card = super().unplay(toHand)
		hist = ("unplay", self.curplayer, card, toHand)
		self._history.append(hist)

	def _undo_unplay(self, player, card, toHand):
		super().play(card, toHand)

	def __setitem__(self, player, hand):
		hist = ("set", player, self[player])
		super().__setitem__(player, hand)
		self._history.append(hist)

	def _undo_set(self, player, pbn):
		super().__setitem__(player, pbn)

	def reset(self):
		"Undo all actions in the undo history"
		while self._history:
			self.undo()

	def checkpoint(self):
		"Clear undo history so that `reset` will bring the deal back to this state"
		self._history = []

	def undo(self):
		"Undo the previous action, if it changed the current deal"
		if self._history:
			action, *args = self._history.pop()
			getattr(self, f"_undo_{action}")(*args)
			return True
		else:
			return False


"""
Provides the `InteractiveDeal` class which keeps an undo stack whenever its 
state is modified, making it easier to interact with the deal. 
"""

__all__ = ["InteractiveDeal"]

from endplay.interact.interactivedeal import InteractiveDeal


"""
Miscellaneous functions relating to play history
"""

from __future__ import annotations

__all__ = [
	"trick_winner", "total_tricks", "tricks_to_result", 
	"result_to_tricks", "linearise_play"]

from endplay.types import Card, Player, Denom, Rank
from collections.abc import Iterable
from more_itertools import grouper

def trick_winner(trick: Iterable[Card], first: Player, trump: Denom):
	"Calculate the winner of a trick"
	winner, topcard = first, trick[0]
	for i in range(1, 4):
		if trick[i].suit == topcard.suit:
			if trick[i].rank > topcard.rank:
				winner, topcard = first.next(i), trick[i]
		elif trick[i].suit == trump:
			winner, topcard = first.next(i), trick[i]
	return winner

def total_tricks(play: Iterable[Card], trump: Denom):
	"""
	Calculate the total number of tricks made by the side
	initially on lead in `play` with the given trump suit
	"""
	first = Player.north
	tricks = 0
	for trick in grouper(iterable=play, n=4):
		winner = trick_winner(trick, first, trump)
		if winner in [Player.north, Player.south]:
			tricks += 1
	return tricks

def tricks_to_result(tricks: int, level: int):
	"""
	Convert tricks made to a result, e.g. 8 tricks
	in a 4-level contract becomes -2
	"""
	return tricks - (level + 6)

def result_to_tricks(result: int, level: int):
	"""
	Convert a result to tricks made, e.g. +1 in a
	3 level contract becomes 10
	"""
	return 6 + level + result

def linearise_play(
		table: list[Card], 
		first: Player, 
		trump: Denom,
		pad_value: Card = Card(suit=Denom.nt, rank=Rank.R2))-> list[str]:
	"""
	Convert a table-style play history to a linear play history.
	PBN record play history is in blocks of four tricks always
	starting with the same player, but many double dummy solving
	algorithms rely on a linear play history where cards are recorded
	in the order they are played. 

	Some data formats include the ability to write cards with an unimportant
	value, and which are required for the table structure. The `pad_value` 
	parameter should be set to a sentinel value (usually a card with nt suit)
	to be recognised
	"""
	play = []
	winner = first
	for row in table:
		# rotate the trick so that the winner is first
		rots = first.turns_to(winner)
		trick = row[rots:] + row[:rots]
		play += trick
		winner = trick_winner(trick, winner, trump)
	# The last trick may have dummy cards (signified by denom=Denom.nt) to
	# fill out the whole trick if there was a claim; if so remove those now
	while play and play[-1] == pad_value:
		play = play[:-1]
	return play

def tabularise_play(
	play: list[Card], 
	first: Player, 
	trump: Denom,
	*,
	pad_value: Card = Card(suit=Denom.nt, rank=Rank.R2)):
	"""
	Convert a linear play sequence into a table, where the first element in each row
	is the card played by the player `first`. 
	If the number of tricks in the play sequence is not not divisible by four, then the
	table will be padded out with `pad_value`.
	"""
	table = []
	winner = first
	for trick in grouper(iterable=play, n=4, fillvalue=pad_value):
		new_winner = trick_winner(trick, winner, trump)
		# rotate the trick so that first is first
		rots = winner.turns_to(first)
		trick = trick[rots:] + trick[:rots]
		table.append(trick)
		winner = new_winner
	return table
	





from __future__ import annotations

__all__ = ["Player"]

from enum import IntEnum
from endplay.types.vul import Vul
from collections.abc import Iterator, Iterable

class Player(IntEnum):
	"Encoding for player seat"
	north	= 0
	east	= 1
	south	= 2
	west	= 3

	@staticmethod
	def find(name: str) -> 'Player':
		"Convert a string into a Player object"
		try:
			if name == "":
				raise ValueError
			return Player("NESW".index(name[0].upper()))
		except ValueError:
			raise ValueError(f"could not convert '{name}' to Player")

	@staticmethod
	def from_lin(n: int) -> 'Player':
		"""
		Convert a BBO LIN representation of a player into a Player object.
		The conversion is determined by 1=S, 2=W, 3=N, 4=E
		"""
		if n < 1 or n > 4:
			raise ValueError(f"n must be in range 1 <= n <= 4")
		return [Player.south, Player.west, Player.north, Player.east][n - 1]

	def to_lin(self) -> int:
		return [3,4,1,2][self]

	@staticmethod
	def from_board(n: int) -> 'Player':
		"""
		Return the player who is the dealer of the corresponding board
		"""
		return Player.west.next(n)

	def is_vul(self, vul: Vul) -> bool:
		return \
			vul == Vul.both or \
			(vul == Vul.ns and self in [Player.north, Player.south]) or \
			(vul == Vul.ew and self in [Player.east, Player.west])

	def enumerate(self, iterable: Iterable, step: int = 1) -> Iterator:
		"""
		Return an iterator whose `next` method returns a tuple
		containing a Player (starting from this player) and the
		values obtained from iterating over iterable, where the
		player increments each time by `step` rotations clockwise
		"""
		player = self
		for item in iterable:
			yield player, item
			player = player.next(step)

	def iter_from(self) -> Iterator['Player']:
		"""
		Iterate over all four players, clockwise, starting from
		this player

		:return: An iterator over all four players in play order
		"""
		for i in range(4):
			yield self.next(i)

	@staticmethod
	def iter_order(order: str) -> Iterator['Player']:
		"""
		Iterate over a sequence of players in a given order

		:param order: The specified order as a four-character string or list of strings
		:return: An iterator over the players in the specified order
		"""
		yield from (Player.find(c) for c in order)

	def turns_to(self, other: 'Player') -> int:
		"Return the number of positions clockwise `other` is from `self`"
		distance = int(other) - int(self)
		if distance < 0:
			distance += 4
		return distance

	def prev(self, n: int = 1) -> 'Player':
		":return: The player who is n places right of the current player"
		return Player((self-n)  % 4)

	def next(self, n: int = 1) -> 'Player':
		":return: The player who is n places left of the current player"
		return Player((self + n) % 4)

	@property
	def lho(self) -> 'Player':
		":return: The player on the current player's left"
		return self.next(1)
	@property
	def partner(self) -> 'Player':
		":return: The player opposite the current player"
		return Player((self + 2) % 4)
	@property
	def rho(self) -> 'Player':
		":return: The player to the current player's right"
		return self.prev(1)

	@property
	def abbr(self) -> str:
		":return: A single character string for the player name"
		return self.name[0].upper()



"""
Actions class for producing LaTeX output.
"""

__all__ = ["LaTeXActions"]

from endplay.dealer.actions.base import BaseActions
from endplay.types import Denom, Player
from io import StringIO
import endplay.stats as stats
import matplotlib.pyplot as plt

class LaTeXActions(BaseActions):
	def __init__(self, deals, stream, board_numbers):
		super().__init__(deals, stream, board_numbers, "tex")

	def print(self, *players):
		exclude = [p for p in Player if p not in players]
		self.write(r"\noindent ")
		for deal in self.deals:
			self.write(r"\resizebox{0.33\textwidth}{!}{" + deal.to_LaTeX(exclude=exclude) + "}")

	def printpbn(self):
		for deal in self.deals:
			self.write(str(deal))

	def printcompact(self, expr = None):
		if expr is None:
			expr = lambda deal: ""
		for deal in self.deals:
			self.write(r"\begin{tabular}{l | l | l | l | l | l}")
			self.write(r"& \textbf{North} & \textbf{East} & \textbf{South} & \textbf{West} & \textbf{Value} \\ \hline")
			self.write(r"$\spadesuit$ &", " & ".join(str(deal[p][Denom.spades]) for p in Player), r"& \\ ")
			self.write(r"$\heartsuit$ &", " & ".join(str(deal[p][Denom.hearts]) for p in Player), "&", expr(deal), r"\\ ")
			self.write(r"$\diamondsuit$ &", " & ".join(str(deal[p][Denom.diamonds]) for p in Player), r"& \\ ")
			self.write(r"$\clubsuit$ &", " & ".join(str(deal[p][Denom.clubs]) for p in Player), r"& \\ ")
			self.write(r"\end{tabular} \\ ")

	def printoneline(self, expr = None):
		for deal in self.deals:
			for player in Player:
				self.write(player.abbr, end=":")
				for denom in Denom.suits():
					self.write("$\\" + denom.name + "uit$" + str(deal[player][denom]), end=' ')
			if expr is not None:
				self.write(f" [{expr(deal)}]", end=' ')
			self.write(r"\\ ")

	def printes(self, *objs):
		for deal in self.deals:
			for obj in objs:
				if callable(obj):
					self.write(obj(deal), end='')
				else:
					self.write(obj, end='')

	def average(self, expr, s = None):
		if s:
			self.write(s, end="")
		self.write(stats.average(self.deals, expr))

	@staticmethod
	def mpl_init_pgf():
		import matplotlib
		matplotlib.use("pgf")
		matplotlib.rcParams.update({
			"pgf.texsystem": "pdflatex",
			'font.family': 'serif',
			'text.usetex': True,
			'pgf.rcfonts': False,
		})

	def frequency1d(self, expr, lb, ub, s = None):
		LaTeXActions.mpl_init_pgf()
		hist = stats.frequency(self.deals, expr, lb, ub)
		fig, ax = plt.subplots()
		ax.bar(list(range(lb, ub+1)), hist)
		if s: ax.set_title(s)
		f = StringIO()
		fig.savefig(f, format="pgf")
		self.write(f.getvalue())

	def frequency2d(self, ex1, lb1, ub1, ex2, lb2, ub2, s = None):
		LaTeXActions.mpl_init_pgf()
		hist = stats.cofrequency(self.deals, ex1, ex2, lb1, ub1, lb2, ub2)
		fig, ax = plt.subplots()
		m = ax.matshow(hist)
		fig.colorbar(m)
		ax.set_xticks(list(range(1 + ub2 - lb2)))
		ax.set_xticklabels([str(i) for i in range(lb2, ub2+1)])
		ax.set_yticks(list(range(1 + ub1 - lb1)))
		ax.set_yticklabels([str(i) for i in range(lb1, ub1+1)])
		if s: ax.set_title(s)
		f = StringIO()
		fig.delaxes(fig.get_axes()[1]) # fixme: currently don't support colorbar in latex
		fig.savefig(f, format="pgf")
		self.write(f.getvalue())


from __future__ import annotations

__all__ = [ "Bid", "ContractBid", "PenaltyBid" ]

from endplay.types.denom import Denom
from endplay.types.penalty import Penalty
from typing import Optional, Union, NoReturn

class Bid:
	"""
	Base class representing an auction call. This class provides a convenience
	constructor from a string, but upon construction will automatically donwcast
	its type to one of :class:`ContractBid` or :class:`PenaltyBid` depending on the type of the
	call.

	:ivar alertable: Flag indicating whether the bid is alertable
	:vartype alertable: bool
	:ivar announcement: String transcription of the announcement for this bid
	:vartype announcement: Optional[str]
	"""
	def __init__(self, name: str, alertable: bool = False, announcement: Optional[str] = None):
		try:
			penalty = Penalty.find(name)
			object.__setattr__(self, "__class__", PenaltyBid)
			PenaltyBid.__init__(self, penalty, alertable, announcement) # type: ignore
		except ValueError:
			level, denom = int(name[0]), Denom.find(name[1])
			object.__setattr__(self, "__class__", ContractBid)
			ContractBid.__init__(self, level, denom, alertable, announcement) # type: ignore

class ContractBid(Bid):
	"""
	Class representing a call that names a contract, i.e. has a level and strain

	:ivar level: The level of the call (between 1 and 7)
	:vartype level: int
	:ivar denom: The strain of the call
	:vartype strain: Denom
	:ivar alertable: Flag indicating whether the bid is alertable
	:vartype alertable: bool
	:ivar announcement: String transcription of the announcement for this bid
	:vartype announcement: Optional[str]
	"""
	def __init__(self,
		level: int,
		denom: Denom,
		alertable: bool = False,
		announcement: Optional[str] = None):
		self.level = level
		self.denom = denom
		self.alertable = alertable
		self.announcement = announcement

	def __repr__(self):
		return f"ContractBid(denom={self.denom!r}, level={self.level!r}, alertable={self.alertable!r}, announcement={self.announcement!r})"

	def __str__(self):
		return f"{self.level}{self.denom.abbr}" + ("!" if self.alertable else "") + ("*" if self.announcement else "")

class PenaltyBid(Bid):
	"""
	Class representing a call that does not name a contract, i.e. pass, double or redouble

	:ivar penalty: The type of the call
	:vartype penalty: Penalty
	:ivar alertable: Flag indicating whether the bid is alertable
	:vartype alertable: bool
	:ivar announcement: String transcription of the announcement for this bid
	:vartype announcement: Optional[str]
	"""
	def __init__(self,
		penalty: Penalty,
		alertable: bool = False,
		announcement: Optional[str] = None):
		self.penalty = penalty
		self.alertable = alertable
		self.announcement = announcement

	def __repr__(self):
		return f"PenaltyBid(penalty={self.penalty!r}, alertable={self.alertable!r}, announcement={self.announcement!r})"

	def __str__(self):
		return (self.penalty.abbr or "p").upper() + ("!" if self.alertable else "") + ("*" if self.announcement else "")



"""
Functions for evaluating bridge hands using a variety of different metrics.
"""

from __future__ import annotations

__all__ = [
	"standard_hcp_scale", "bergen_hcp_scale", "hcp", "shortage_nofit_dist_scale",
	"shortage_fit_dist_scale", "length_dist_scale", "mixed_fit_dist_scale",
	"mixed_nofit_dist_scale", "dist_points", "total_points",
	"top_honours", "losers", "cccc", "quality", "controls", "rule_of_n",
	"exact_shape", "shape", "major_shape", "minor_shape", "is_balanced",
	"is_semibalanced", "is_minor_semibalanced", "is_single_suited",
	"is_two_suited", "is_three_suited"]

from typing import Union, Optional
from collections.abc import Iterable
from endplay.types import Hand, SuitHolding, Card, Rank, AlternateRank, Denom

# Points for each rank in order [A, K, Q, J, ...]
standard_hcp_scale: list[float] = [4, 3, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]
"""Classic HCP scale with A=4, K=3, Q=2, J=1"""
bergen_hcp_scale: list[float] = [4.5, 3, 1.5, 0.75, 0.25, 0, 0, 0, 0, 0, 0, 0, 0]
"""Bergen HCP scale with A=4.5, K=3, Q=1.5, J=0.75, T=0.25"""

def hcp(
	obj: Union[Iterable[Union[Card, Rank, AlternateRank]], Card, Rank, AlternateRank],
	scale: list[float] = standard_hcp_scale) -> float:
	"""
	Return the high card points of a hand or suit holding using a given scale. Two
	scales are predefined: `standard_hcp_scale` which uses the traditional 4321
	point scale, and `bergen_hcp_scale` which uses the Bergen scale.

	:param obj: A (possibly iterable of) card or rank type to evaluate
	:param scale: A list of 13 numbers which assign points to each rank in descending
		order, e.g. [4,3,2,1,0,...] is used for the standard HCP scale where an ace is
		worth 4 points, a king 3 etc...
	"""
	if len(scale) < 13:
		scale += [0] * (13 - len(scale))
	if isinstance(obj, Card):
		return scale[14 - obj.rank.to_alternate()]
	elif isinstance(obj, Rank):
		return scale[14 - obj.to_alternate()]
	elif isinstance(obj, AlternateRank):
		return scale[14 - obj]
	else:
		return sum(hcp(card, scale) for card in obj)

# Points for each suit length in order [0/void, 1/singleton, 2/doubleton, 3, ...]
shortage_nofit_dist_scale: list[float] = [3, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
"""Distribution point scale with void=3, singleton=2, doubleton=1"""
shortage_fit_dist_scale: list[float] = [5, 3, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
"""Distribution point scale when a trump fit is found with void=5, singleton=3, doubleton=1"""
length_dist_scale: list[float] = [0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
"""Distribution point scale which awards a point for each extra card in a suit over 4"""
mixed_nofit_dist_scale: list[float] = [3, 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
"""Combined scale using `shortage_nofit_dist_scale` and `length_dist_scale`"""
mixed_fit_dist_scale: list[float] = [5, 3, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
"""Combined scale using `shortage_fit_dist_scale` and `length_dist_scale`"""

def dist_points(obj: Union[Hand, SuitHolding], scale: list[float] = shortage_nofit_dist_scale, exclude: Iterable[Denom] = []) -> float:
	"""
	Return the extra points for distribution for a hand or suit holding using the given scale.
	Five scales are predefined:

	* `shortage_nofit_dist_scale`: Standard evaluation of shortage points if no trump fit is agreed,
	  void=3, singleton=2, doubleton=1

	* `shortage_fit_dist_scale`: Standard evaluation of shortage points if trump fit is agreed,
	  void=5, singleton=3, doubleton=1

	* `length_dist_scale`: Standard evaluation of length points, one point for each extra card in a
	  suit with more than four cards in it, i.e. 5 cards=1, 6 cards=2, etc...

	:param obj: Hand or suit holding to evaluate
	:param scale: A list of 14 numbers which assign points to each suit length in ascending order,
		e.g. [3,2,1,0,...] is used for the standard shortage points where void=3, singleton=2 etc...
	:param exclude: List of suits to be excluded from distribution calculations if a `Hand` object
		is passed, in order to e.g. exclude the trump suit from the calculation
	"""
	if len(scale) < 14:
		scale += [0] * (14 - len(scale))
	if isinstance(obj, SuitHolding):
		return scale[len(obj)]
	return sum(dist_points(obj[suit], scale) if suit not in exclude else 0 for suit in Denom.suits())

def total_points(
	obj: Union[Hand, SuitHolding],
	hcp_scale: list[float] = standard_hcp_scale,
	dist_scale: list[float] = shortage_nofit_dist_scale,
	trump: Optional[Denom] = None,
	protect_honours: bool = False):
	"""
	Return the sum of high card points and distribution points in a hand.

	:param obj: Hand or suit holding to evaluate
	:param hcp_scale: The HCP scale to use, as for the `hcp` function
	:param dist_scale: The distribution points scale, as for the `dist_scale` function
	:param trump: If set then distribution points for the given suit are excluded from
		the calculation
	:param protect_honours: If `True`, then honours must be protected for their HCP
		to be included, i.e. king singleton, queen doubleton etc do not contribute to
		total points
	"""
	if isinstance(obj, SuitHolding):
		points = dist_points(obj, dist_scale, [trump] if trump is not None else [])
		if protect_honours:
			obj = obj.copy()
			ncards = len(obj)
			if ncards < 4: obj.remove(Rank.RJ)
			if ncards < 3: obj.remove(Rank.RQ)
			if ncards < 2: obj.remove(Rank.RK)
		return hcp(obj, hcp_scale) + points
	else:
		return sum(total_points(obj[suit], hcp_scale, dist_scale, trump, protect_honours) for suit in Denom.suits())

def top_honours(hand: Union[Hand, SuitHolding], lowest_honour: Union[Rank, int] = Rank.RJ) -> int:
	"""
	Return the number of top honors in a suit

	:param suit: The suit holding to evaluate
	:param lowest_honour: The lowest rank to be treated as an honour, or an
		integer for how many top cards are honours
	"""
	if not isinstance(lowest_honour, Rank):
		lowest_honour = AlternateRank(15 - lowest_honour).to_standard()
	if isinstance(hand, SuitHolding):
		return sum(1 for rank in hand if rank >= lowest_honour)
	else:
		return sum(1 for card in hand if card.rank >= lowest_honour)

def losers(hand: Union[Hand, SuitHolding]) -> int:
	"""
	Returns the number of losers in a hand. The number of losers in each
	suit is calculated as: (a) 3 or more cards in a suit: 3 - 1 per AKQ
	(b) Doubleton: AK=0, Ax/Kx=1, else 2 (c) Singleton: A=0, else 1 (d) Void: 0
	"""
	if isinstance(hand, SuitHolding):
		a, k, q = Rank.RA in hand, Rank.RK in hand, Rank.RQ in hand
		if len(hand) == 0: return 0
		if len(hand) == 1: return 1 - a
		if len(hand) == 2: return 2 - a - k
		return 3 - a - k - q
	else:
		return sum(losers(hand[suit]) for suit in Denom.suits())

def cccc(hand: Union[Hand, SuitHolding]) -> float:
	"""
	Uses the Kaplan four cs algorithm from the October 1982 issue of Bridge world
	magazine. This implementation is based on the interpretation of the algorithm
	described on http://www.rpbridge.net/8j19.htm
	"""
	if isinstance(hand, Hand):
		return sum(cccc(hand[suit]) for suit in Denom.suits())
	l = len(hand)
	score = 0
	# 1-5: Count point values A=4,K=3,Q=2,J=1,T=0.5
	if Rank.RA in hand: score += 4
	if Rank.RK in hand: score += 3
	if Rank.RQ in hand: score += 2
	if Rank.RJ in hand: score += 1
	if Rank.RT in hand: score += 0.5
	# 6-7: if between 2-6 cards, count points for supported T and 9
	if l >= 2 and l <= 6:
		if (Rank.RT in hand) and ((Rank.RJ in hand) or top_honours(hand, 4) >= 2):
			score += 0.5
		if (Rank.R9 in hand) and ((Rank.R8 in hand or Rank.RT in hand) or top_honours(hand, 5) == 2):
			score += 0.5
	# 8: if between 4-6 cards, count 9 without any touching cards and exactly three higher honours
	if l >= 4 and l <= 6:
		if (Rank.R9 in hand) and not (Rank.R8 in hand and Rank.RT in hand) and top_honours(hand, 4) == 3:
			score += 0.5
	# 9: 7+ cards missing queen or jack (or both)
	if l >= 7 and (Rank.RQ not in hand or Rank.RJ not in hand):
		score += 1
	# 10: 8+ cards missing queen
	if l >= 8 and Rank.RQ not in hand:
		score += 1
	# 11: 9+ cards missing queen and jack
	if l >= 9 and (Rank.RQ not in hand and Rank.RJ not in hand):
		score += 1
	# 12: Multiply current total by suit length, divide by 10, then continue
	score *= l / 10
	# 13: Add 3 for an ace
	if Rank.RA in hand:
		score += 3
	# 14-15: Add 2 for guarded king and 0.5 for king singleton
	if Rank.RK in hand:
		if l > 1: score += 2
		else: score += 0.5
	# 16-19: Add values for queen
	if Rank.RQ in hand:
		if l >= 3:
			# 16-17: 3+ cards, +1 if has higher honour else +0.75
			if Rank.RK in hand or Rank.RA in hand:
				score += 1
			else:
				score += 0.75
		elif l == 2:
			# 18-19: doubleton, +0.5 if has higher honour else +0.25
			if Rank.RK in hand or Rank.RA in hand:
				score += 0.5
			else:
				score += 0.25
	# 20-21: Jack with two higher honours=0.5, one higher honour=0.25
	if Rank.RJ in hand:
		hh = top_honours(hand, 3)
		if hh == 2:
			score += 0.5
		elif hh == 1:
			score += 0.25
	# 22-23: Ten with two higher honours=0.25, T9 with one higher honour=0.25
	if Rank.RT in hand:
		hh = top_honours(hand, 4)
		if hh == 2:
			score += 0.25
		if Rank.R9 in hand and hh == 1:
			score += 0.25
	# 24-26: void=3, singleton=2, doubleton=1
	if l == 0:
		score += 3
	elif l == 1:
		score += 2
	elif l == 2:
		score += 1
	return score

def quality(hand: SuitHolding) -> float:
	"Uses the quality algorithm (from the October 1982 issue of Bridge World magazine)"
	raise NotImplementedError

def controls(hand: Union[Iterable, Card, Rank]) -> int:
	"Return the number of controls in a sequence, using A=2 and K=1"
	if isinstance(hand, Rank):
		if hand == Rank.RA: return 2
		elif hand == Rank.RK: return 1
		else: return 0
	elif isinstance(hand, Card):
		return controls(hand.rank)
	else:
		return sum(controls(card) for card in hand)

def rule_of_n(hand: Hand) -> float:
	"Returns the hcp of the hand added to the length of the two longest suits"
	s = shape(hand)
	return hcp(hand) + s[0] + s[1]

def exact_shape(hand: Hand) -> list[int]:
	"Return the shape of a hand as a list starting from spades, e.g. (5, 2, 3, 3)"
	return [len(hand[suit]) for suit in Denom.suits()]

def shape(hand: Hand) -> list[int]:
	"""
	Return the shape of a hand as a list from longest to shortest, e.g. a 3424 would
	return (4, 4, 3, 2)
	"""
	return sorted(exact_shape(hand), reverse=True)

def major_shape(hand: Hand) -> list[int]:
	"Return the shape of a hand's major holding from longest to shortest"
	return sorted((len(hand[Denom.spades]), len(hand[Denom.hearts])), reverse=True)

def minor_shape(hand: Hand) -> list[int]:
	"""
	Return the shape of a hand's minor holding from longest to shortest
	"""
	return sorted((len(hand[Denom.diamonds]), len(hand[Denom.clubs])), reverse=True)

def is_balanced(hand: Hand) -> bool:
	"""
	Returns True if the hand shape is 4333, 4432 or 5332
	"""
	s = shape(hand)
	return s in ([4,3,3,3],[4,4,3,2],[5,3,3,2])

def is_semibalanced(hand: Hand) -> bool:
	"""
	Returns True if the hand shape is balanced or 5422
	"""
	s = shape(hand)
	return is_balanced(hand) or s == [5, 4, 2, 2]

def is_minor_semibalanced(hand: Hand) -> bool:
	"""
	Returns True if the hand shape is balanced or contains
	doubletons in both minors
	"""
	return is_balanced(hand) or minor_shape(hand) == [2, 2]

def is_single_suited(hand: Hand, min_length: int = 6, no_side_suit: bool = False) -> bool:
	"""
	Returns True if the hand is singled suited

	:param min_length: The minimum number of cards to hold in the suit
	:param no_side_suit: If True, the hand cannot contain another 4 card suit
	"""
	s = shape(hand)
	if no_side_suit:
		return s[0] >= min_length and s[1] < 4
	else:
		return s[0] >= min_length

def is_two_suited(hand: Hand, strict: bool = False) -> bool:
	"""
	Returns True if the hand contains at least 10 cards in two suits.

	:param strict: Hand must be at least 5-5
	"""
	s = shape(hand)
	if strict:
		return s[0] + s[1] >= 10 and s[1] >= 5
	else:
		return s[0] + s[1] >= 10

def is_three_suited(hand: Hand, strict: bool = False) -> bool:
	"""
	Returns True if the hand has three suits with at least four cards in them.

	:param strict: Only return true if the hand is 4441
	"""
	s = shape(hand)
	if strict:
		return s == [4, 4, 4, 1]
	else:
		return s[0] >= 4 and s[1] >= 4 and s[2] >= 4



"""
Par contract and scoring function
"""

from __future__ import annotations

__all__ = ["ParList", "par"]

from typing import Union
import endplay._dds as _dds
from endplay.types import Deal, Contract, Vul, Player
from endplay.dds.ddtable import calc_dd_table, DDTable
from ctypes import pointer
from collections.abc import Iterable, Iterator

class ParList(Iterable):
	def __init__(self, data: '_dds.parResultsMaster'):
		self._data = data

	@property
	def score(self) -> int:
		":return: The score associated with the par contracts"
		return self._data.score

	def __iter__(self) -> Iterator[Contract]:
		":return: An iterator over all the par contracts"
		for i in range(self._data.number):
			c = self._data.contracts[i]
			if c.seats == 4:
				tmp = _dds.contractType()
				pointer(tmp)[0] = c
				c.seats = 0
				yield Contract(c)
				pointer(tmp)[0] = c
				c.seats = 2
				yield Contract(c)
			elif c.seats == 5:
				tmp = _dds.contractType()
				pointer(tmp)[0] = c
				c.seats = 1
				yield Contract(c)
				pointer(tmp)[0] = c
				c.seats = 3
				yield Contract(c)
			else:
				yield Contract(self._data.contracts[i])

	def __repr__(self) -> str:
		return '<ParList object>'

def par(
	deal: Union[Deal, DDTable], 
	vul: Union[Vul, int], 
	dealer: Player) -> ParList:
	"""
	Calculate the par contract result for the given deal. 

	:param deal: The deal to calculate the par result of. If you have already precomputed a DDTable for
		a deal then you speed up the par calculation by passing that instead of the Deal object
	:param vul: The vulnerability of the deal. If you pass an `int` then this is converted from a board
		number into the vulnerability of that board
	:param dealer: The dealer of the board.
	"""
	if isinstance(deal, Deal):
		dd_table = calc_dd_table(deal)
	else:
		dd_table = deal
	if not isinstance(vul, Vul):
		vul = Vul.from_board(vul)

	par = _dds.parResultsMaster()
	_dds.DealerParBin(dd_table._data, par, dealer, vul)
	return ParList(par)




"""
Analysis functions from the DDS library, which calculate the double dummy
number of tricks available given a play history.
"""

from __future__ import annotations

__all__ = [ 
	"SolvedPlay", "SolvedPlayList", "analyse_start", "analyse_play", 
	"analyse_all_plays", "analyse_all_starts" ]

from typing import Union
from collections.abc import Iterable, Sequence
from endplay.types import Deal, Card
import endplay._dds as _dds

class SolvedPlay(Sequence):
	def __init__(self, data: _dds.solvedPlay):
		self._data = data

	def __len__(self) -> int:
		":return: The number of results in the play"
		return self._data.number

	def __getitem__(self, i: int) -> int:
		":return: The number of tricks that declarer can make after the ith card is played"
		if i < 0:
			i = len(self) - i
		if i < 0 or i >= len(self):
			raise IndexError
		return self._data.tricks[i]

	def __repr__(self) -> str:
		return f'<SolvedPlay object; data={self!s}>'

	def __str__(self) -> str:
		return "(" + ", ".join(str(p) for p in self) + ")"

class SolvedPlayList(Sequence):
	def __init__(self, data: _dds.solvedPlays):
		self._data = data

	def __len__(self) -> int:
		":return: The number of boards in the list"
		return self._data.noOfBoards

	def __getitem__(self, i: int) -> SolvedPlay:
		":return: The solved play at index `i`"
		if i < 0:
			i = len(self) - i
		if i < 0 or i >= len(self):
			raise IndexError
		return SolvedPlay(self._data.solved[i])

	def __repr__(self) -> str:
		return f'<SolvedPlayList; length={len(self)}'

	def __str__(self) -> str:
		return "[" + ", ".join(str(s) for s in self) + "]"

def analyse_start(deal: Deal, declarer_is_first: bool = False) -> int:
	"""
	Calculate the most tricks declarer can make.

	:param deal: The deal to analyse
	:param declarer_is_first: The algorithm assumes that the person who leads is to the left
		of the declarer (as would be the case with the first card led
		to a hand), but to return the result as seen from the leader's
		perspective you can set this to True
	"""
	# Create empty play trace
	playBin = _dds.playTraceBin()
	playBin.number = 0
	# Calculate and return 13-n, as it returns the tricks from the perspective of RHO
	solvedp = _dds.solvedPlay()
	_dds.AnalysePlayBin(deal._data, playBin, solvedp, 0)
	if declarer_is_first:
		return len(deal[deal.first]) - solvedp.tricks[0]
	else:
		return solvedp.tricks[0]

def analyse_play(
	deal: Deal, 
	play: Iterable[Union[Card, str]], 
	declarer_is_first: bool = False) -> SolvedPlay:
	"""
	Calculate a list of double dummy values after each card in `play`
	is played to the hand. This returns `len(play)+1` results, as there
	is also a result before any card has been played
	"""
	# Convert play to playTraceBin
	playBin = _dds.playTraceBin()
	playBin.number = 0
	for i, card in enumerate(play):
		if isinstance(card, str):
			card = Card(card)
		playBin.suit[i] = card.suit
		playBin.rank[i] = card.rank.to_alternate()
		playBin.number += 1

	# Solve and correct trick count if perspective is wrong way round
	solvedp = _dds.solvedPlay()
	_dds.AnalysePlayBin(deal._data, playBin, solvedp, 0)
	if declarer_is_first:
		starting_cards = len(deal[deal.first])
		for i in range(solvedp.number):
			solvedp.tricks[i] = starting_cards - solvedp.tricks[i]
	return SolvedPlay(solvedp)

def analyse_all_starts(
	deals: Iterable[Deal], 
	declarer_is_first: bool = False) -> list[int]:
	"""
	Optimized version of analyse for multiple deals which uses threading to
	speed up the calculation
	"""
	# Convert deals into boards
	bop = _dds.boards()
	bop.noOfBoards = 0
	if declarer_is_first:
		# We need to keep track of how many cards the person on lead starts
		# with, so that we can invert the result
		starting_cards = []
	for i, deal in enumerate(deals):
		if i > _dds.MAXNOOFBOARDS:
			raise RuntimeError(f"Too many boards, maximum is {_dds.MAXNOOFBOARDS}")
		if declarer_is_first:
			starting_cards.append(len(deal[deal.first]))
		bop.deals[i] = deal._data
		bop.target[i] = -1
		bop.solutions[i] = 3
		bop.mode[i] = 0
		bop.noOfBoards += 1
	# Create empty play traces
	plp = _dds.playTracesBin()
	plp.noOfBoards = bop.noOfBoards
	for i in range(plp.noOfBoards):
		plp.plays[i].number = 0

	# Calculate and return 13-n, as it returns the tricks from the perspective of RHO
	solvedp = _dds.solvedPlays()
	_dds.AnalyseAllPlaysBin(bop, plp, solvedp, 0)
	if declarer_is_first:
		return [starting_cards[i] - solvedp.solved[i].tricks[0] for i in range(plp.noOfBoards)]
	else:
		return [solvedp.solved[i].tricks[0] for i in range(plp.noOfBoards)]

def analyse_all_plays(
	deals: Iterable[Deal], 
	plays: Iterable[Iterable[Card]], 
	declarer_is_first: bool = False) -> SolvedPlayList:
	"""
	Optimized version of analyse_play for multiple deals which uses
	threading to speed up the calculation
	"""
	# Convert deals into boards
	bop = _dds.boards()
	bop.noOfBoards = 0
	if declarer_is_first:
		starting_cards = []
	for i, deal in enumerate(deals):
		if i > _dds.MAXNOOFBOARDS:
			raise RuntimeError(f"Too many boards, maximum is {_dds.MAXNOOFBOARDS}")
		if declarer_is_first:
			starting_cards.append(len(deal[deal.first]))
		bop.deals[i] = deal._data
		bop.target[i] = -1
		bop.solutions[i] = 3
		bop.mode[i] = 0
		bop.noOfBoards += 1

	# Convert plays into playTracesBin
	plp = _dds.playTracesBin()
	plp.noOfBoards = 0
	for i, play in enumerate(plays):
		plp.plays[i].number = 0
		for j, card in enumerate(play):
			if isinstance(card, str):
				card = Card(card)
			plp.plays[i].suit[j] = card.suit
			plp.plays[i].rank[j] = card.rank.to_alternate()
			plp.plays[i].number += 1
		plp.noOfBoards += 1

	solvedp = _dds.solvedPlays()
	_dds.AnalyseAllPlaysBin(bop, plp, solvedp, 0)
	if declarer_is_first:
		for i in range(bop.noOfBoards):
			for j in range(solvedp.solved[i].number):
				solvedp.solved[i].tricks[j] = starting_cards[i] - solvedp.solved[i].tricks[j]
	return SolvedPlayList(solvedp)


from __future__ import annotations

__all__ = ["Denom"]

from enum import IntEnum
from collections.abc import Iterator

class Denom(IntEnum):
	"Encoding for suits and contract denomination"
	spades		= 0
	hearts		= 1
	diamonds	= 2
	clubs		= 3
	nt			= 4

	@staticmethod
	def find(name: str) -> 'Denom':
		"Convert a string value into a Denom object"
		try:
			if name == "":
				raise ValueError
			return Denom("SHDCN♠♥♦♣N♤♡♢♧".index(name[0].upper()) % 5)
		except ValueError:
			raise ValueError(f"Could not convert {name} into a Denom object")

	@staticmethod
	def bidorder() -> Iterator['Denom']:
		":return: An iterator over all the denominations in the order they appera in a bidding box"
		yield from [Denom.clubs, Denom.diamonds, Denom.hearts, Denom.spades, Denom.nt]

	@staticmethod
	def suits(reverse: bool = False) -> Iterator['Denom']:
		"""
		Iterate over the suits, excluding notrumps

		:param reverse: If true, return suits in the order clubs -> spades
		:return: An iterator over the four suits
		"""
		r = range(3,-1,-1) if reverse else range(4)
		for i in r:
			yield Denom(i)

	def is_suit(self) -> bool:
		":return: True if the denomination is not notrumps"
		return self in Denom.suits()

	def is_major(self) -> bool:
		":return: True if the denomination is spades or hearts"
		return self in [ Denom.spades, Denom.hearts ]

	def is_minor(self) -> bool:
		":return: True if the denomination is diamonds or clubs"
		return self in [ Denom.diamonds, Denom.clubs ]

	@property
	def abbr(self) -> str:
		":return: A short identifier for the denomination"
		import endplay.config as config
		if self == Denom.nt:
			return "NT"
		elif config.use_unicode:
			return "♠♥♦♣"[self]
		else:
			return "SHDC"[self]



__all__ = ["Card"]

from typing import Optional
from dataclasses import dataclass

from endplay.types.denom import Denom
from endplay.types.rank import Rank

@dataclass(frozen=True)
class Card:
	"""
	Immutabale class representing a card with `suit` and `rank` read-only attributes

	:ivar suit: The suit of the card
	:vartype suit: Denom
	:ivar rank: The rank of the card
	:vartype rank: Rank
	"""
	rank: Rank
	suit: Denom

	def __init__(self, name: Optional[str] = None, *, suit: Optional[Denom] = None, rank: Optional[Rank] = None):
		"""
		Construct a card either from a string name or from Denom and Rank objects

		:param name: The name of the card, e.g. "S9" or "HT"
		:param suit: The suit of the card
		:param rank: The rank of the card
		"""
		if name is not None:
			suit = Denom.find(name[0])
			rank = Rank.find(name[1])

		if not isinstance(suit, Denom):
			raise ValueError("suit must be of type Denom")
		object.__setattr__(self, "suit", suit)

		if not isinstance(rank, Rank):
			raise ValueError("rank must be of type Rank")
		object.__setattr__(self, "rank", rank)

	def __str__(self) -> str:
		return f"{self.suit.abbr}{self.rank.abbr}"



"""
Frontends contain a :class:`Deal` object as state, and provide an
`interact` method as their entry point which allows interactive
manipulation of the deal.
"""

from endplay.interact.frontends.cmd import CmdFrontend
from endplay.interact.frontends.curses import CursesFrontend
from endplay.interact.frontends.tk import TkFrontend


__all__ = ["Vul"]

from enum import IntEnum

class Vul(IntEnum):
	"Encodes the vulnerability"
	none	= 0
	both	= 1
	ns		= 2
	ew		= 3

	@staticmethod
	def find(name: str) -> 'Vul':
		"Convert a string into a Vul object"
		if name.lower() in [ "n", "s", "ns" ]:
			return Vul.ns
		elif name.lower() in [ "e", "w", "ew" ]:
			return Vul.ew
		elif name.lower() in [ "all", "both", "a", "b" ]:
			return Vul.both
		elif name.lower() in [ "", "none", "-", "luv" ]:
			return Vul.none
		else:
			raise ValueError(f"could not convert '{name}' to Vul")

	@staticmethod
	def from_lin(s: str) -> 'Vul':
		"""
		Convert a BBO LIN string of vulnerability into a Vul object.
		The conversion is determined by o=none, e=ew, n=ns, b=both
		"""
		try:
			if s == "":
				raise ValueError
			return Vul("obne".index(s.lower()))
		except ValueError:
			raise ValueError(f"invalid lin vulnerability '{s}': must be one of o, b, n, e")

	def to_lin(self) -> str:
		return "obne"[self]

	@staticmethod
	def from_board(board_no: int) -> 'Vul':
		":return: The vulnerability of the specified board"
		i = board_no - 1
		shift = (i // 4) % 4
		return [Vul.none ,Vul.ns, Vul.ew, Vul.both][(i + shift) % 4]


	@property
	def abbr(self) -> str:
		"A short string representation of the vulnerability"
		if self == Vul.none:
			return "-"
		if self == Vul.both:
			return "A"
		return self.name.upper()



"""
Functionality for generating bridge deals. This submodule provides an API via
:func:`generate_deals` as well as an executable module which can be run with
`python3 -m endplay.dealer` and which aims to emulate the behaviour of 
Hans van Staveren's original dealer program
"""

__all__ = ['run_script', 'generate_deal', 'generate_deals']

from endplay.dealer.generate import generate_deal, generate_deals
from endplay.dealer.constraint import ConstraintInterpreter
from endplay.dealer.runscript import run_script



"Parser for BridgeBase LIN files"

from __future__ import annotations

__all__ = [ "LINEncodeError", "dump", "dumps", "load", "loads" ]

from typing import TextIO
from more_itertools import chunked
from endplay.types import Deal, Bid, ContractBid, Vul, Card, Player, Board, Contract
from endplay.utils.play import result_to_tricks, tricks_to_result, total_tricks
from endplay.utils.escape import unescape_suits, escape_suits
from endplay.config import suppress_unicode

class LINDecoder:
	"""
	Class providing functionality for reading the LIN file format
	"""

	def __init__(self):
		pass

	def parse_line(self, line: str) -> Board:
		deal = None
		auction = []
		play = []
		board_num = None
		dealer = None
		vul = None
		contract = None
		claimed = False
		info = {}

		elems = line.split("|")
		pairs = [(a, b) for a, b in zip(elems[::2], elems[1::2])]
		for key, value in pairs:
			if key == "st":
				continue
			elif key == "pn":
				# Player names are comma separated starting from south
				info["South"], info["West"], info["North"], info["East"] = value.split(",")
			elif key == "md":
				# Marks deal, starts with dealer then comma separated hands
				dealer = Player.from_lin(int(value[0]))
				deal = Deal.from_lin(value, complete_deal = True)
			elif key == "sv":
				# Marks vulnerability
				vul = Vul.from_lin(value)
			elif key == "ah":
				# Marks board number in format 'Board N'
				board_num = int(value[5:])
			elif key == "mb":
				# Marks a bid
				auction.append(Bid(value, value[-1:] == "!"))
			elif key == "an":
				# Marks an alert for the previous bid
				auction[-1].announcement = unescape_suits(value)
			elif key == "pc":
				# Marks a card in the play section
				play.append(Card(value))
			elif key == "pg":
				# Signifies end of play
				continue
			elif key == "mc":
				# Marks that tricks were claimed
				claimed = True
				if contract is None:
					contract = Contract.from_auction(dealer, auction)
					contract.result = tricks_to_result(int(value), contract.level)
			else:
				info[key] = value
		# ensure there is a contract if there was an auction
		if contract is None and (dealer is not None and len(auction) > 0):
			contract = Contract.from_auction(dealer, auction)
			if play:
				tricks = total_tricks(play, contract.denom)
				contract.result = tricks_to_result(int(tricks), contract.level)
		# make sure 'first' and 'trump' in deal are set correctly if there is
		# a contract
		if contract is not None:
			deal.first = contract.declarer.lho
			deal.trump = contract.denom
		return Board(deal, auction, play, board_num, vul=vul, dealer=dealer, contract=contract, claimed=claimed, **info)

	def parse_string(self, lin: str) -> list[Board]:
		boards = []
		for line in lin.splitlines():
			boards.append(self.parse_line(line))
		return boards

	def parse_file(self, f: TextIO) -> list[Board]:
		boards = []
		for line in f:
			boards.append(self.parse_line(line))
		return boards

class LINEncodeError(ValueError):
    pass

class LINEncoder:
	def __init__(self):
		pass

	def serialise_board(self, board: Board) -> str:
		# Early error checking
		if board.dealer is None:
			raise LINEncodeError("Cannot encode board without `dealer` set")
		if board.deal is None:
			raise LINEncodeError("Cannot encode board without `deal` set")
		with suppress_unicode():
			lin = ""
			s, w, n, e = board.info.south, board.info.west, board.info.north, board.info.east
			lin += f"pn|{s or ''},{w or ''},{n or ''},{e or ''}|"
			lin += "st||"
			lin += f"md|{board.deal.to_lin(board.dealer)}|"
			lin += "rh||"
			if board.board_num is not None:
				lin += f"ah|Board {board.board_num}|"
			if board.vul is not None:
				lin += f"sv|{board.vul.to_lin()}|"
			for bid in board.auction:
				lin += "mb|"
				if isinstance(bid, ContractBid):
					lin += f"{bid.level}{bid.denom.abbr[0]}" + ("!" if bid.alertable else "") + "|"
				else:
					lin += bid.penalty.name[0] + ("!" if bid.alertable else "") + "|"
				if bid.announcement is not None:
					lin += f"an|{escape_suits(bid.announcement)}|"
			lin += "pg||"
			for trick in chunked(board.play, 4):
				for card in trick:
					lin += f"pc|{card}|"
				if len(trick) == 4:
					lin += "pg||"
			if board.claimed:
				tricks = result_to_tricks(board.contract.result, board.contract.level)
				lin += f"mc|{tricks}|"
		return lin

def load(fp: TextIO) -> list[Board]:
    "Read a LIN file object into an array of :class:`Board` objects"
    parser = LINDecoder()
    return parser.parse_file(fp)

def loads(s: str) -> list[Board]:
    "Read a LIN string into a n arrya of :class:`Board` objects"
    parser = LINDecoder()
    return parser.parse_string(s)

def dump(boards: list[Board], fp: TextIO) -> None:
	"Seralize a list of :class:`Board` objects to a LIN file"
	parser = LINEncoder()
	for board in boards:
		line = parser.serialise_board(board)
		fp.write(line + "\n")
	
def dumps(boards: list[Board]) -> str:
	"Serialize a list of :class:`Board` objects to a LIN string"
	parser = LINEncoder()
	lines = []
	for board in boards:
		lines.append(parser.serialise_board(board))
	return "\n".join(lines) + "\n"



"""
A curses interface to many of the features of endplay including 
dealing hands and performing double-dummy analysis.
"""

__all__ = ["CursesFrontend"]

import curses
from curses.textpad import Textbox

class CursesFrontend:
	def __init__(self):
		raise NotImplementedError
		self.command_history = []
		self.ps1 = "> "

	def _main(self, stdscr):
		rows, cols = stdscr.getmaxyx()
		
		# Deal window
		deal_win = curses.newwin(0, 0, 12, 39)

		# DDTable window

		# DDTricks window

		stdscr.hline(12, 0, curses.ACS_HLINE, cols)

		# Command history window
		hist_height = rows - 15
		hist_width = cols
		hist_win = curses.newwin(hist_height, hist_width, 13, cols)

		stdscr.hline(rows - 2, 0, curses.ACS_HLINE, cols)

		# Input window
		stdscr.addstr(rows - 1, 0, self.ps1)
		edit_win = curses.newwin(1, cols - len(self.ps1), rows - 1, len(self.ps1))

		while True:
			stdscr.noutrefresh()
			deal_win.noutrefresh()
			hist_win.noutrefresh()
			edit_win.noutrefresh()
			curses.doupdate()

			csl_in = Textbox(edit_win)
			cmd = csl_in.edit()
			self.command_history.append(cmd)
			for i, h in enumerate(self.command_history[-hist_height::-1]):
				hist_win.addstr(hist_height - i - 1, 0, h)


	def interact(self):
		curses.wrapper(self._main)



"""
Miscellaneous input and output routines
"""

from __future__ import annotations

__all__ = ["pprint_auction"]

from endplay.types import Player, Bid
from typing import TextIO
import sys
from collections.abc import Iterable
from more_itertools import chunked

def pprint_auction(
		first: Player, 
		auction: Iterable[Bid], 
		*,
		include_announcements=False,
		stream: TextIO = sys.stdout):
	for player in first.iter_from():
		print(player.abbr.ljust(5), end="", file=stream)
	print(file=stream)
	for row in chunked(auction, 4):
		for bid in row:
			print(str(bid).ljust(5), end="", file=stream)
		print(file=stream)
	if include_announcements:
		print(file=stream)
		for bid in auction:
			if bid.announcement:
				print(f"{bid}: {bid.announcement.strip()}", file=stream)


"""
Parser for PBN files
"""

from __future__ import annotations

__all__ = ["PBNDecoder", "load", "loads", "dump", "dumps"]

import re
from enum import Enum
from io import StringIO
from itertools import chain
from typing import TextIO, Union

from more_itertools import chunked

from endplay.config import suppress_unicode
from endplay.types import (Bid, Board, Card, Contract, ContractBid, Deal,
                           Denom, Player, Rank, Vul)
from endplay.utils.play import (linearise_play, result_to_tricks,
                                tabularise_play, tricks_to_result)


class PBNDecodeError(ValueError):
	def __init__(self, msg, line, lineno):
		super().__init__(msg)
		self.line = line
		self.lineno = lineno

class PBNDecoder:
	# Precompile some regular expressions
	re_lcomment = re.compile(r"\s*;\s*(.*?)")
	re_bcomment_line = re.compile(r"\s*{\s*(.*?)}")
	re_bcomment_begin = re.compile(r"\s*{\s*(.*)")
	re_bcomment_end = re.compile(r"(.*)}")
	re_ignore = re.compile(r"%.*")
	re_fileformat = re.compile(r"%\s*(EXPORT|IMPORT)", re.IGNORECASE)
	re_pbnversion = re.compile(r"%\s*PBN (\d+)\.(\d+)", re.IGNORECASE)
	re_metatag = re.compile(r"%\s*([\w-]+):\s*(.*?)")
	re_tagpair = re.compile(r"\[(\w+)\s+\"(.*?)\"\](\s*[;{]?.*)")
	re_note = re.compile(r"(\d+):(.*)")
	re_colname = re.compile(r"([+-]?)(\w+)(?:\\(\d+)([LR]?))?", re.IGNORECASE)

	class State(Enum):
		NONE = 0
		META = 1
		DATA = 2
		COMMENTBLOCK = 3

	def __init__(self):
		#warnings.warn("endplay.parsers.PBNDecoder is now deprecated and subject to be replaced with a different interface at any time")
		self.metadata = {}

	def _get_comment(self, text: str, iscont: bool):
		"""
		Returns the comment contained in text, and a flag which
		is set to True if the comment requires continuation or
		False otherwise
		"""
		if iscont:
			m = PBNDecoder.re_bcomment_end.match(text)
			if m:
				return (m.group(1), False)
			return (text, True)
		else:
			# Match a single line comment (begins with a semicolon)
			m = PBNDecoder.re_lcomment.match(text)
			if m:
				return (m.group(1), False)

			# Match a block comment that ends on the same line (e.g. { xyz })
			m = PBNDecoder.re_bcomment_line.match(text)
			if m:
				return (m.group(1), False)

			# Match a block comment that begins on the line but is continued (e.g. { xy)
			m = PBNDecoder.re_bcomment_begin.match(text)
			if m:
				return (m.group(1), True)

			# No comment found
			return None

	def _tags_to_board(self):
		if len(self.curtags) == 0:
			return
		board = Board()
		declarer = None
		tricks = None
		# Loop through keys, if the key has a special variable defined in the Board class,
		# then attach the value to that variable. Otherwise, add it as a tag-value(-data)
		# entry in the `info` dict
		for raw_key, fields in self.curtags.items():
			key = raw_key.lower()
			if fields["value"] == "#":
				fields = self.prevtags[raw_key]
			value = fields["value"]
			if key == "board":
				board.board_num = int(value)
			elif key == "vulnerable":
				board.vul = Vul.find(value) if value else None
			elif key == "deal":
				board.deal = Deal(value)
			elif key == "declarer":
				declarer = Player.find(value) if value else None
			elif key == "contract":
				board.contract = Contract(value or "Pass")
			elif key == "result":
				tricks = int(value or "0")
			elif key == "auction":
				# Requires converting; iterate over the bids and check for
				# special values and deal with them appropriately, otherise
				# just convert to a bid
				board.dealer = Player.find(value) if value else None
				flattened_auction = [b for row in fields["data"] for b in row]
				for bid in flattened_auction:
					if bid.lower() == "ap":
						# All pass, add three passes
						for _ in range(3):
							board.auction.append(Bid("pass"))
					elif bid.startswith("=") and bid.endswith("="):
						# Note; grab the reference and attach it to the previous
						# bid's alert
						idx = bid[1:-1]
						board.auction[-1].alertable = True
						board.auction[-1].announcement = self.notes.get(idx, "Note not found")
					elif bid in "-*+":
						# either the auction has finished, or it is a unknown call. Either 
						# way, parsing the auction is finished for us.
						break
					else:
						board.auction.append(Bid(bid))
			elif key == "play":
				first = Player.find(value)
				table = []
				# Convert card names to Card objects. Ignore notes (=N=), and use
				# a card with NT suit to represent an unknown/unimportant/padding card
				# Then linearise the result so that we get a flat array of cards in 
				# play order
				for raw_row in fields["data"]:
					if len(raw_row) == 1 and raw_row[0] == "*":
						board.claimed = True
						break
					row = []
					for card in raw_row:
						if card.startswith("=") and card.endswith("="):
							pass # ignore play annotations for now
						elif card in "+-":
							row.append(Card(suit=Denom.nt, rank=Rank.R2))
						elif card == "*":
							board.claimed = True
							row.append(Card(suit=Denom.nt, rank=Rank.R2))
						else:
							row.append(Card(card))
					table.append(row)
				board.play = linearise_play(table, first, board.contract.denom)
			else:
				# By default, add to the info section
				if len(fields) == 1:
					board.info[raw_key] = value
				else:
					board.info[raw_key] = dict(headers=fields["value"], rows=fields["data"])

		# housekeeping: if declarer/result are manually specified
		# then add to the contract
		if board.contract is not None:
			if declarer is not None:
				board.contract.declarer = declarer
			if tricks is not None:
				board.contract.result = tricks_to_result(tricks, board.contract.level)
			# set first and trump on deal
			if board.deal is not None:
				board.deal.first = board.contract.declarer.lho
				board.deal.trump = board.contract.denom
		return board

	def _parse_meta(self, curline):
		# Match against PBN version (e.g. % PBN 2.1)
		m = PBNDecoder.re_pbnversion.match(curline)
		if m:
			self.metadata["version"] = { "major": int(m.group(1)), "minor": int(m.group(2)) }
			return False
		# Match against IMPORT/EXPORT (e.g. % EXPORT)
		m = PBNDecoder.re_fileformat.match(curline)
		if m:
			self.metadata["format"] = m.group(1).upper()
			return False
		# Match against some other metadata (e.g. % Creator: Joe Bloggs)
		m = PBNDecoder.re_metatag.match(curline)
		if m:
			self.metadata[m.group(1)] = m.group(2)
			return False
		# Some other comment line we can ignore
		if PBNDecoder.re_ignore.match(curline):
			return False
		# No match found, metadata section complete
		self.state = PBNDecoder.State.NONE
		return True

	def _parse_conttable(self, curline):
		# Revert to PBNDecoder.State.NONE if a tag or empty line is encountered
		if PBNDecoder.re_tagpair.match(curline) or curline == "":
			self.state = PBNDecoder.State.NONE
		# Split the line by whitespace add add the elements to the tag value
		else:
			self.curtags[self.curtag]["data"] += [curline.split()]
		return True

	def _parse_none(self, curline):
		# Empty line, start new game (or ignore if the current game is empty)
		if curline == "":
			if self.curtags:
				board = self._tags_to_board()
				self.boards.append(board)
				self.prevtags, self.curtags = self.curtags, {}
			return False
		# Comment line, ignore
		if PBNDecoder.re_ignore.match(curline):
			return False
		# Tag pair
		m = PBNDecoder.re_tagpair.match(curline)
		if m:
			self.curtag = m.group(1)
			# Ignore tag if repeated
			if self.curtag in self.curtags:
				return False
			# Get the comment attached to the line (if any)
			if m.group(3):
				_, needcont = self._get_comment(m.group(3), False)
				if needcont:
					self.state = PBNDecoder.State.COMMENTBLOCK
			# Add the tag to the current game, entering into State.DATA if it is a 
			# section that expects data (play, auction or *table)
			if self.curtag.lower() == "play" or self.curtag.lower() == "auction":
				self.curtags[self.curtag] = { "value": m.group(2), "data": [] }
				self.state = PBNDecoder.State.DATA
			elif self.curtag.lower().endswith("table") and self.curtag.lower() != "table":
				colnames = []
				for colname in m.group(2).split(";"):
					cm = PBNDecoder.re_colname.match(colname)
					if cm:
						if (not cm.group(1)) and (not cm.group(3)) and (not cm.group(4)):
							colnames += [cm.group(2)]
						else:
							colnames += [{
								"ordering": cm.group(1) or None,
								"name": cm.group(2),
								"minwidth": cm.group(3) or None,
								"alignment": cm.group(4) or None
							}]
					else:
						raise PBNDecodeError("Could not parse column name", curline, self.lineno)
				self.curtags[self.curtag] = { "value": colnames, "data": [] }
				self.state = PBNDecoder.State.DATA
			elif self.curtag.lower() == "note":
				# Keep a separate list of notes which will be attached to their relevant items
				idx, note = m.group(2).split(":", maxsplit=1)
				self.notes[idx] = note
			else:
				# Any other tag gets created as a tag-value pair
				self.curtags[self.curtag] = { "value": m.group(2) }
			return False
		else:
			raise PBNDecodeError("Expected a tag", curline, self.lineno)

	def _parse_commentblock(self, curline):
		_, needcont = self._get_comment(curline, True)
		if not needcont:
			self.state = PBNDecoder.State.NONE
		return True

	def parse_file(self, f: TextIO) -> list[Board]:
		"Parse a PBN file"

		self.boards = []
		self.prevtags = {}
		self.curtags = {}
		self.notes = {}
		self.state = PBNDecoder.State.META
		self.curtag = None
		self.lineno = 0

		# Loop over lines, keeping track of what type of line we are expecting.
		# If the line isn't consumed, change the state and fallthrough to a 
		# method which can consume the line. Always append a blank line to the
		# end of the file input to ensure that the last board is processed
		for curline in chain(f.readlines(), [""]):
			self.lineno += 1
			curline = curline.strip()

			if curline.startswith("%") and self.state != PBNDecoder.State.META:
				continue

			if self.state == PBNDecoder.State.META:	
				if not self._parse_meta(curline):
					continue
			if self.state == PBNDecoder.State.DATA:
				if not self._parse_conttable(curline):
					continue
			if self.state == PBNDecoder.State.NONE:
				if not self._parse_none(curline):
					continue
			if self.state == PBNDecoder.State.COMMENTBLOCK:
				if not self._parse_commentblock(curline):
					continue

		return self.boards

class PBNEncoder:
	def __init__(self):
		self.tags = []
		self.pbn_version = "2.1"

	def _create_tag(self, key, value, data=None):
		key = key[:1].upper() + key[1:]
		tagpair = f'[{key} "{value}"]'
		if data is not None:
			for line in data:
				tagpair += "\n" + " ".join(str(cell) for cell in line)
		self.tags.append(tagpair + "\n")

	def _serialize_columns(self, headers: list[Union[dict, str]]) -> str:
		colnames = []
		for col in headers:
			if isinstance(col, str):
				colname = col
			else:
				colname = ""
				if col["ordering"] is not None:
					colname += col["ordering"]
				colname += col["name"]
				if col["minwidth"] is not None:
					colname += "\\" + col["minwidth"]
					if col["alignment"] is not None:
						colname += col["alignment"]
			colnames.append(colname)
		return ";".join(colnames)

	def _justify_row(self, headers: list[dict], row: list[str]) -> list[str]:
		res = []
		for header, cell in zip(headers, row):
			if isinstance(header, str):
				minwidth = 0
				alignment = "L"
			else:
				minwidth = int(header["minwidth"] or 0)
				alignment = header["alignment"] or "L"
			if alignment == "L":
				res.append(cell.ljust(minwidth))
			else:
				res.append(cell.rjust(minwidth))
		return res

	def export_board(self, board: Board):
		# Split the board info into mandatory and supplementary tags
		event = "?"
		site = "?"
		date = "?"
		west = "?"
		north = "?"
		east = "?"
		south = "?"
		scoring = "?"
		supplementary = {}
		for raw_key, value in board.info.items():
			key = raw_key.lower()
			if key == "event":
				event = value
			elif key == "site":
				site = value
			elif key == "date":
				date = value
			elif key == "west":
				west = value
			elif key == "north":
				north = value
			elif key == "east":
				east = value
			elif key == "south":
				south = value
			elif key == "scoring":
				scoring = value
			elif key.endswith("table") and key != "table":
				headers = self._serialize_columns(value["headers"])
				rows = [self._justify_row(value["headers"], row) for row in value["rows"]]
				supplementary[raw_key] = Board.Info(value=headers, data=rows)
			elif isinstance(value, str):
				supplementary[raw_key] = Board.Info(value=value)
			else:
				supplementary[raw_key] = value

		with suppress_unicode():
			# Print mandatory tags
			self.tags = []
			self._create_tag("Event", event)
			self._create_tag("Site", site)
			self._create_tag("Date", date)
			self._create_tag("Board", board.board_num)
			self._create_tag("West", west)
			self._create_tag("North", north)
			self._create_tag("East", east)
			self._create_tag("South", south)
			if board.dealer is None:
				self._create_tag("Dealer", "?")
			else:
				self._create_tag("Dealer", board.dealer.abbr)
			if board.vul is None:
				self._create_tag("Vulnerable", "?")
			else:
				v = ["None", "All", "NS", "EW"][board.vul]
				self._create_tag("Vulnerable", v)
			self._create_tag("Deal", board.deal.to_pbn())
			self._create_tag("Scoring", scoring)
			if board.contract is None:
				self._create_tag("Declarer", "?")
			else:
				self._create_tag("Declarer", board.contract.declarer.abbr)
			if board.contract is None:
				self._create_tag("Contract", "?")
				self._create_tag("Result", "?")
			else:
				c = board.contract
				if c.is_passout():
					cstr = "Pass"
					res = ""
				else:
					cstr = f"{c.level}{c.denom.abbr}{c.penalty.abbr}"
					res = result_to_tricks(board.contract.result, board.contract.level)
				self._create_tag("Contract", cstr)
				self._create_tag("Result", res)

			# Print play and auction
			notes = {}
			if board.auction:
				table = []
				for row in chunked(board.auction, 4):
					tablerow = []
					for bid in row:
						if isinstance(bid, ContractBid):
							tablerow.append(f"{bid.level}{bid.denom.abbr}")
						else:
							tablerow.append(bid.penalty.abbr.upper() or "Pass")
						if bid.announcement:
							idx = len(notes) + 1
							notes[idx] = bid.announcement
							tablerow.append(f"={idx}=")
					table.append(tablerow)
				self._create_tag("Auction", board.dealer.abbr, table)

			if board.play:
				pad_value = Card(suit=Denom.nt, rank=Rank.R2)
				table = tabularise_play(
					board.play, 
					board.contract.declarer.lho,
					board.contract.denom)
				string_table = []
				for row in table:
					string_row = []
					for card in row:
						if card == pad_value:
							string_row.append("- ")
						else:
							string_row.append(str(card))
					string_table.append(string_row)
				if board.claimed:
					string_table.append(["*"])

				self._create_tag("Play", board.contract.declarer.lho.abbr, string_table)

			# Print notes
			for idx, note in notes.items():
				self._create_tag("Note", f"{idx}:{note}")

			# Print supplementary tags
			for key, value in supplementary.items():
				self._create_tag(key, value.value, value.data)


	def export_file(self, boards: list[Board], fp: TextIO) -> None:
		"Export current data into a file object"
		fp.write(f"% PBN {self.pbn_version}\n")
		fp.write("% EXPORT\n\n")
		for board in boards:
			self.export_board(board)
			fp.writelines(self.tags)
			fp.write("\n")


def load(fp: TextIO) -> list[Board]:
	"Read a PBN file object into a list of :class:`Board` objects"
	parser = PBNDecoder()
	return parser.parse_file(fp)

def loads(s: str) -> list[Board]:
	"Read a PBN string into a list of :class:`Board` objects"
	parser = PBNDecoder()
	sp = StringIO(s)
	return parser.parse_file(sp)

def dump(boards: list[Board], fp: TextIO) -> None:
	"Serialize a list of :class:`Board` objects to a PBN file"
	parser = PBNEncoder()
	parser.export_file(boards, fp)

def dumps(boards: list[Board]) -> str:
	"Serialize a list of :class:`Board` objects to a PBN string"
	parser = PBNEncoder()
	sp = StringIO()
	parser.export_file(boards, sp)
	return sp.getvalue()



__all__ = ["Rank", "AlternateRank"]

from enum import IntEnum

class Rank(IntEnum):
	"""
	Encodes the rank of a suit. The standard values use
	powers of two, however some internal functions use
	an alternative encoding AlternateRank using the
	values 2-14.
	"""
	R2 = 0x0004
	R3 = 0x0008
	R4 = 0x0010
	R5 = 0x0020
	R6 = 0x0040
	R7 = 0x0080
	R8 = 0x0100
	R9 = 0x0200
	RT = 0x0400
	RJ = 0x0800
	RQ = 0x1000
	RK = 0x2000
	RA = 0x4000

	@property
	def abbr(self) -> str:
		return self.name[1]

	@staticmethod
	def find(value: str) -> 'Rank':
		try:
			return Rank[f"R{value.upper()}"]
		except KeyError:
			raise ValueError(f"could not convert '{value}' to Rank")

	def to_alternate(self) -> 'AlternateRank':
		# Calculate integer log2
		x, y = self.value, 0
		while True:
			x >>= 1
			if x == 0:
				break
			y += 1
		return AlternateRank(y)

class AlternateRank(IntEnum):
	"""
	Encodes the rank of a suit using the values 2-14. Used
	for internal functions, for APIs use the Rank class.
	"""
	R2 =  2
	R3 =  3
	R4 =  4
	R5 =  5
	R6 =  6
	R7 =  7
	R8 =  8
	R9 =  9
	RT = 10
	RJ = 11
	RQ = 12
	RK = 13
	RA = 14

	@property
	def abbr(self) -> str:
		return self.name[1]

	@staticmethod
	def find(value: str) -> 'AlternateRank':
		try:
			return AlternateRank[f"R{value.upper()}"]
		except KeyError:
			raise ValueError(f"could not convert '{value}' to AlternateRank")

	def to_standard(self) -> Rank:
		# Calculate integer power of 2
		return Rank(2 << (self.value-1))



"Parser for endplay formatted JSON files"

from __future__ import annotations

__all__ = ["JSONEncoder", "JSONDecoder", "dump", "dumps", "load", "loads"]

import json as _json
from enum import IntEnum
from collections.abc import Collection, Mapping
from endplay.types import ContractBid, PenaltyBid, Denom, Penalty, \
	Card, Rank, Contract, Player, Deal, Vul, Board, Hand, SuitHolding

def _check_keys(keys: set, mandatory: set, optional: set = set()):
	if not mandatory.issubset(keys):
		return False
	return (keys - mandatory).issubset(optional)

def _object_hook(d: dict):
	keys = set(d.keys())
	if _check_keys(keys, 
		set(["level", "denom"]), 
		set(["alertable", "announcement"])):
		return ContractBid(
			d.get("level"), Denom.find(d.get("denom")), 
			d.get("alertable", False), d.get("announcement"))
	elif _check_keys(keys, 
		set(["penalty"]), 
		set(["alertable", "announcement"])):
		return PenaltyBid(
			Penalty.find(d.get("penalty")), 
			d.get("alertable", False), d.get("announcement"))
	elif _check_keys(keys, set(["suit", "rank"])):
		return Card(
			suit=Denom.find(d.get("suit")), 
			rank=Rank.find(d.get("rank")))
	elif _check_keys(keys, 
		set(["level", "denom", "declarer"]), 
		set(["penalty", "result"])):
		return Contract(
			level=d.get("level"),
			denom=Denom.find(d.get("denom")),
			declarer=Player.find(d.get("declarer")),
			penalty=Penalty.find(d.get("penalty", "pass")),
			result=d.get("result", 0))
	elif _check_keys(keys,
		set(["north", "south", "east", "west"]),
		set(["first", "trump", "curtrick"])):
		deal = Deal(
			first=Player.find(d.get("first", "north")),
			trump=Denom.find(d.get("trump", "nt")))
		for hand in ["north", "south", "east", "west"]:
			for card in d.get(hand, []):
				deal[Player.find(hand)].add(card)
		for card in d.get("curtrick", []):
			deal.play(card)
		return deal
	elif _check_keys(keys,
		set(["deal", "auction", "play", "board_num"]),
		set(["vul", "dealer", "contract", "claimed", "info"])):
		board = Board(
			deal=d.get("deal"),
			auction=d.get("auction"),
			play=d.get("play"),
			board_num=d.get("board_num"),)
		if "vul" in d:
			board.vul = Vul.find(d.get("vul"))
		if "dealer" in d:
			board.dealer = Player.find(d.get("dealer"))
		if "contract" in d:
			board.contract = d.get("contract")
		if "claimed" in d:
			board.claimed = d.get("claimed")
		if "info" in d:
			for key, val in d.get("info").items():
				board.info[key] = val
		return board
	else:
		return d 

class JSONDecoder(_json.JSONDecoder):
	"""
	Class providing functionality for reading the JSON file format
	"""
	def __init__(self, 
		*, object_hook=None, parse_float=None, parse_int=None,
		parse_constant=None, strict=True, object_pairs_hook=None):
		if object_hook is not None or object_pairs_hook is not None:
			raise RuntimeError("object_hook and object_pairs_hook currently not supported")
		super().__init__(
			object_hook = _object_hook,
			parse_float = parse_float,
			parse_int = parse_int,
			parse_constant = parse_constant,
			strict = strict
		)

def _preprocess_intenum(o):
    if isinstance(o, IntEnum):
        return {'__class__': o.__class__.__name__,
                '__value__': (o.value,)}
    if isinstance(o, dict):
        return {k: _preprocess_intenum(v) for k, v in o.iteritems()}
    if isinstance(o, (list, tuple)):
        return [_preprocess_intenum(v) for v in o]
    return o

class JSONEncoder(_json.JSONEncoder):
	"""
	Class providing functionality for writing to the JSON file format
	"""
	def iterencode(self, o, _one_shot=False):
		return super().iterencode(_preprocess_intenum(o), _one_shot)

	def default(self, o):
		if isinstance(o, ContractBid):
			res = {
				"level": o.level,
				"denom": o.denom.name,
				"alertable": o.alertable,
			}
			if o.announcement:
				res["announcement"] = o.announcement
			return res
		elif isinstance(o, PenaltyBid):
			res = {
				"penalty": "pass" if o.penalty is Penalty.passed else o.penalty.name[:-1],
				"alertable": o.alertable
			}
			if o.announcement:
				res["announcement"] = o.announcement
			return res
		elif isinstance(o, Board):
			res = {
				"deal": o.deal,
				"auction": o.auction,
				"play": o.play,
				"board_num": o.board_num,
				"info": dict(o.info)
			}
			if o.vul is not None:
				res["vul"] = o.vul.name
			if o.dealer is not None:
				res["dealer"] = o.dealer.name
			if o.contract is not None:
				res["contract"] = o.contract
			if o.claimed is not None:
				res["claimed"] = o.claimed
			return res
		elif isinstance(o, Card):
			return {
				"suit": o.suit.name,
				"rank": o.rank.name[1]
			}
		elif isinstance(o, Contract):
			return {
				"level": o.level,
				"denom": o.denom.name,
				"declarer": o.declarer.name,
				"penalty": o.penalty.abbr,
				"result": o.result
			}
		elif isinstance(o, Deal):
			return {
				"north": list(o.north),
				"east": list(o.east),
				"south": list(o.south),
				"west": list(o.west),
				"first": o.first.name,
				"trump": o.trump.name,
				"curtrick": list(o.curtrick)
			}
		elif isinstance(o, Hand):
			return list(o)
		elif isinstance(o, SuitHolding):
			return [r.name for r in o] 
		else:
			return super().default(self, o)

def dump(obj, fp, 
	*, skipkeys=False, ensure_ascii=True, check_circular=True, 
	allow_nan=True, indent=None, separators=None, default=None, 
	sort_keys=False, **kw):
	return _json.dump(obj, fp,
		skipkeys=skipkeys,
		ensure_ascii=ensure_ascii,
		check_circular=check_circular,
		allow_nan=allow_nan,
		cls=JSONEncoder,
		indent=indent,
		separators=separators,
		default=default,
		sort_keys=sort_keys,
		**kw)

def dumps(obj, 
	*, skipkeys=False, ensure_ascii=True, check_circular=True, 
	allow_nan=True, indent=None, separators=None, default=None, 
	sort_keys=False, **kw):
	return _json.dumps(obj,
		skipkeys=skipkeys,
		ensure_ascii=ensure_ascii,
		check_circular=check_circular,
		allow_nan=allow_nan,
		cls=JSONEncoder,
		indent=indent,
		separators=separators,
		default=default,
		sort_keys=sort_keys,
		**kw)

def load(fp, *, 
	object_hook=None, parse_float=None, parse_int=None, 
	parse_constant=None, object_pairs_hook=None, **kw):
	return _json.load(fp,
		cls=JSONDecoder,
		object_hook=object_hook,
		parse_float=parse_float,
		parse_int=parse_int,
		parse_constant=parse_constant,
		object_pairs_hook=object_pairs_hook,
		**kw
	)

def loads(s, *, 
	object_hook=None, parse_float=None, parse_int=None, 
	parse_constant=None, object_pairs_hook=None, **kw):
	return _json.loads(s, 
		cls=JSONDecoder,
		object_hook=object_hook,
		parse_float=parse_float,
		parse_int=parse_int,
		parse_constant=parse_constant,
		object_pairs_hook=object_pairs_hook,
		**kw)


"""
Generic routines for generating deals based on a set of constraints
"""

from __future__ import annotations

__all__ = ['generate_deal', 'generate_deals']

from endplay.dealer.constraint import ConstraintInterpreter, Expr
from endplay.types import Deal, Card, Denom, Player, Rank
from numpy.random import RandomState # guaranteed to be stable for numpy>=1.16
from typing import Optional, Union
from collections.abc import Iterator
import warnings
from tqdm import trange

class DealNotGeneratedError(RuntimeError):
	pass

class DealNotGeneratedWarning(Warning):
	pass

class InconsistentSwappingAlgorithmWarning(Warning):
	pass

# This module could be better implemented as a C extension, with
# a function like the following assigning bits directly to the 
# underlying datatype instead of constructing a list of Python
# objects, shuffling it and then slicing the list
#
#void complete_deal(uint16_t** deal)
#{
#	// Initialize probabilities and flatten dealt cards
#	uint16_t predeal[4] = { 0, 0, 0, 0 };
#	int probs[4] = { 13, 26, 39, 52 };
#	for (int hand = 0; hand < 4; ++hand) {
#		for (int suit = 0; suit < 4; ++suit) {
#			int count = popcount(deal[hand][suit]);
#			predeal[suit] |= deal[hand][suit];
#			for (int i = hand; i < 4; ++i)
#				probs[i] -= count;
#		}
#	}
#
#	// Deal remaining cards
#	for (int rank = 0; rank < 13; ++rank) {
#		uint16_t rank_bit = 1u << rank;
#		for (int suit = 0; suit < 4; ++suit) {
#			// Skip if card already dealt
#			if (predeal[suit] & rank_bit)
#				continue;
#			// Pick a random number, and whichever range in probs
#			// the number falls into we add the card to the corresponding
#			// hand. We then decrement all elements in probs starting with
#			// this hand. This reduces the probablility of a card being
#			// dealt to the hand we just added a card too, and then adjusts
#			// the other ranges to account for the fact that there is one
#			// fewer card to deal.
#			int k = rand() % probs[3];
#			if (k < probs[0]) {
#				deal[0][suit] |= rank_bit;
#				--probs[0], --probs[1], --probs[2], --probs[3];
#			}
#			else if (k < probs[1]) {
#				deal[1][suit] |= rank_bit;
#				--probs[1], --probs[2], --probs[3];
#			}
#			else if (k < probs[2]) {
#				deal[2][suit] |= rank_bit;
#				--probs[2], --probs[3];
#			}
#			else {
#				deal[3][suit] |= rank_bit;
#				--probs[3];
#			}
#		}
#	}
#}

def generate_deal(
	*constraints: Union[Expr, str], 
	predeal: Deal = Deal(), 
	swapping: int = 0, 
	seed: Optional[int] = None,
	max_attempts: int = 1000000, 
	env: dict = {}) -> Deal:
	"""
	Generates a random deal satisfying the constraints, giving 13 cards to each player.
	The constraints should be supplied as functions taking a whole deal and returning
	a boolean, for example `lambda d: hcp(d.north) > 10`, or as expressions compatible 
	with the dealer expression syntax (see https://www.bridgebase.com/tools/dealer/Manual/input.html#expr)

	:param constraints: Constraints, as callables or strings
	:param predeal: A :class:`Deal` object which may be partially filled with cards; these will not
		be shuffled, allowing you to specify that players should have particular holdings.
	:param swapping: An integer representing the type of swapping algorithm to use, either
		0 (no swapping), 2 (swapping EW) or 3 (swapping EWS)
	:param seed: The number to seed the random generator with. A `numpy` random generator is
		used which is guaranteed to be stable between Python releases.
	:param max_attempts: Maximum number of shuffles to perform when finding a deal which
		matches the constraints. Set to -1 for infinite
	:param env: A dictionary of the environment used when evaluating constraints
	"""
	# We just call generate deals and return the first deal that is yielded. The parameters
	# are the same and so are just forwarded onwards, with the exception that we always set
	# exhaustion_is_error to True (as it really *is* an error to not generate a single hand)
	# and we set show_progress to False as this would show all zeros right until the one hand
	# we want to produce is produced.
	deals = generate_deals(
		*constraints, 
		predeal = predeal,
		swapping = swapping, 
		show_progress = False,
		produce = 1,
		seed = seed, 
		max_attempts = max_attempts, 
		env = env,
		strict = True)
	return next(deals)

def generate_deals(
	*constraints: Union[Expr, str], 
	predeal: Deal = Deal(), 
	swapping: int = 0, 
	show_progress: bool = False,
	produce: int = 40,
	seed: Optional[int] = None,
	max_attempts: int = 1000000, 
	env: dict = {},
	strict: bool = False) -> Iterator[Deal]:
	"""
	Generates `produce` random deals satisfying the constraints which should
	be given as for :func:`generate_deal`. `produce` and `max_attemps` are upper limits,
	the first to be reached terminates the program

	:param constraints: Constraints, as callables or strings
	:param predeal: A :class:`Deal` object which may be partially filled with cards; these will not
		be shuffled, allowing you to specify that players should have particular holdings.
	:param swapping: An integer representing the type of swapping algorithm to use, either
		0 (no swapping), 2 (swapping EW) or 3 (swapping EWS)
	:param show_progress: If True, a progress bar is displayed with information on how many hands
		have been generated so far
	:param produce: The total number of hands satisfying the constraints to find.
	:param seed: The number to seed the random generator with. A `numpy` random generator is
		used which is guaranteed to be stable between Python releases.
	:param max_attempts: Maximum number of shuffles to perform when finding a deal which
		matches the constraints. Set to -1 for infinite
	:param env: A dictionary of the environment used when evaluating constraints
	:param strict: If True, a `RuntimeError` is raised if `max_attempts` is reached before
		`produce` hands are produced. Otherwise, a warning is generated
	"""
	if swapping == 2 and (len(predeal.west) > 0 or len(predeal.east) > 0):
		warnings.warn("2-way swapping is incompatible with E/W predealt, output may be unexpected", InconsistentSwappingAlgorithmWarning)
	elif swapping == 3 and (len(predeal.west) > 0 or len(predeal.east) > 0 or len(predeal.south) > 0):
		warnings.warn("3-way swapping is incompatible with E/W/S predealt, output may be unexpected", InconsistentSwappingAlgorithmWarning)

	rs = RandomState(seed)

	ci = ConstraintInterpreter()
	for name, val in env.items():
		ci.set_env(name, val)
	constraints = [ci.lambdify(c) if not callable(c) else c for c in constraints]
	cards = set(Card(suit=denom, rank=rank) for denom in Denom.suits() for rank in Rank)
	cards = list(cards.difference(predeal.to_hand()))
	split = [sum([13 - len(hand) for _, hand in predeal][:i]) for i in range(5)]
	generated = 0
	if show_progress:
		prange = trange(produce, desc="Produced", unit="deals")
	else:
		prange = range(produce)
	for p in prange:
		if show_progress and p > 0:
			prange.set_postfix({"success": f"{100*p/generated:.2f}%"})
		produced = False
		while not produced:
			if generated == max_attempts:
				if show_progress:
					prange.close()
				message = f"Only {p} out of {produce} hands were generated before max_attempts (set to {max_attempts}) was reached"
				if strict:
					raise DealNotGeneratedError(message)
				else:
					warnings.warn(message, DealNotGeneratedWarning)
					return generated
			generated += 1
			rs.shuffle(cards)
			deal = predeal.copy()
			for i, player in enumerate(Player):
				deal[player].extend(cards[split[i]:split[i+1]])
			for perm in _generate_swaps(deal, swapping):
				if all(c(perm) for c in constraints):
					yield perm
					produced = True
					break
	return generated

def _generate_swaps(deal: Deal, swapping: int):
	if swapping == 0:
		yield deal
	elif swapping == 2:
		yield deal
		deal.swap(1, 3)
		yield deal
	elif swapping == 3:
		for _ in range(3):
			deal.swap(1, 2)
			yield deal
			deal.swap(2, 3)
			yield deal
	else:
		raise ValueError(f"Invalid swapping value {swapping} used")



from __future__ import annotations

__all__ = ["SuitHolding"]

import ctypes
from endplay.types.rank import Rank, AlternateRank
from typing import Optional, Union
from collections.abc import Iterable, Iterator

class SuitHolding:
	def __init__(self, data: Union[ctypes.c_uint * 4, str] = "", idx: Optional[int] = None):
		"""
		Construct a suit holding

		:param data: A PBN string representation of the holding or a reference to a _dds object
		"""
		if isinstance(data, str):
			self._data = (ctypes.c_uint * 4)(0,0,0,0)
			self._idx = 0
			for rank in data:
				self.add(rank)
		else:
			if idx is None:
				raise ValueError("No index given to SuitHolding")
			self._idx = idx
			self._data = data

	def __copy__(self) -> 'SuitHolding':
		return SuitHolding((ctypes.c_uint * 4).from_buffer_copy(self._data), self._idx)

	def copy(self) -> 'SuitHolding':
		return self.__copy__()

	def add(self, rank: Union[Rank, str]) -> bool:
		"""
		Add a rank to the suit holding

		:param rank: The rank to add
		:return: False if the rank was already in the holding, True otherwise
		"""
		if isinstance(rank, str):
			rank = Rank.find(rank)
		if isinstance(rank, AlternateRank):
			rank = rank.to_standard()
		if rank in self:
			return False
		self._data[self._idx] |= rank.value
		return True

	def extend(self, ranks: Iterable[Rank]) -> int:
		"""
		Add multiple ranks to the suit holding

		:parm ranks: An iterable of the ranks to add
		:return: The number of ranks successfully added
		"""
		return sum(self.add(rank) for rank in ranks)

	def remove(self, rank: Rank) -> bool:
		"""
		Remove a rank from the suit holding

		:param rank: The rank to remove
		:return: False if the rank wasn't in the holding, True otherwise
		"""
		if isinstance(rank, str):
			rank = Rank.find(rank)
		if isinstance(rank, AlternateRank):
			rank = rank.to_standard()
		if not rank in self:
			return False
		self._data[self._idx] &= ~rank.value
		return True

	def clear(self) -> None:
		"""
		Removes all cards from the holding
		"""
		self._data[self._idx] = 0

	@staticmethod
	def from_pbn(pbn: str) -> 'SuitHolding':
		"""
		Construct a SuitHolding object from a PBN string
		"""
		return SuitHolding(pbn)

	def to_pbn(self) -> str:
		"Create a PBN representation of the suit holding"
		return "".join(rank.abbr for rank in self)

	def __eq__(self, other: object) -> bool:
		if not isinstance(other, SuitHolding):
			return NotImplemented
		return self._data[self._idx] == other._data[other._idx]

	def __contains__(self, rank: Rank) -> bool:
		if isinstance(rank, str):
			rank = Rank.find(rank)
		if isinstance(rank, AlternateRank):
			rank = rank.to_standard()
		return bool(self._data[self._idx] & rank.value)

	def __iter__(self) -> Iterator[Rank]:
		for rank in reversed(Rank):
			if rank in self: yield rank

	def __reversed__(self) -> Iterator[Rank]:
		for rank in Rank:
			if rank in self: yield rank

	def __len__(self) -> int:
		n = self._data[self._idx]
		b = bin(n)
		return b.count('1')

	def __str__(self) -> str:
		return self.to_pbn()

	def __repr__(self) -> str:
		return f'SuitHolding("{self!s}")'



"""
A shell-like CLI interface to many of the features of endplay including 
dealing hands and performing double-dummy analysis.
"""

import endplay.config as config
from endplay.types import Player, Denom
from endplay.interact.interactivedeal import InteractiveDeal
from endplay.interact.frontends import CmdFrontend, CursesFrontend, TkFrontend
import argparse

def main():
	parser = argparse.ArgumentParser(description=__doc__)
	parser.add_argument("-a", "--ascii", action='store_true', help="Turn off unicode output")
	parser.add_argument("-n", "--north", help="The cards in the north hand as a PBN string")
	parser.add_argument("-e", "--east", help="The cards in the east hand as a PBN string")
	parser.add_argument("-s", "--south", help="The cards in the south hand as a PBN string")
	parser.add_argument("-w", "--west", help="The cards in the west hand as a PBN string")
	parser.add_argument("-d", "--deal", help="The deal as a PBN string (note this must be quoted)")
	parser.add_argument("-f", "--first", help="The player initially on lead")
	parser.add_argument("-t", "--trump", help="The trump suit")
	parser.add_argument("-x", "--frontend", default='cmd', choices=["cmd", "curses", "tk"], help="Which frontend engine to use, can be one of 'cmd', 'curses' or 'tk'. Defaults to 'cmd'.")
	parser.add_argument("-v", "--verbose", action="store_true", help="Make error messages more verbose (helpful for debugging).")
	args = parser.parse_args()

	config.use_unicode = not args.ascii

	d = InteractiveDeal(args.deal)
	if args.first: d.first = Player.find(args.first)
	if args.trump: d.trump = Denom.find(args.trump)
	if args.north: d.north = args.north
	if args.east: d.east = args.east
	if args.south: d.south = args.south
	if args.west: d.west = args.west

	if args.frontend == "cmd":
		frontend = CmdFrontend(d, verbose_errors=args.verbose)
	elif args.frontend == "curses":
		frontend = CursesFrontend(d)
	elif args.frontend == "tk":
		frontend = TkFrontend(d)
	else:
		print("Unknown frontend '{args.frontend}' specified")
		exit(1)

	frontend.interact()

if __name__ == "__main__":
	main()


"""
Engine for converting dealer-syntax expressions into evaluatable
Python functions. You should not need to use this directly unless
you are developing your own functions, as most functions accept
strings as well as functions and automatically compile them using
the machinery in this module.
"""

__all__ = [ "ConstraintInterpreter" ]

import re
from typing import Callable, Any, Union
from endplay.parsers.dealer import Node, DealerParser
from endplay.types import Deal, Denom
from endplay.evaluate import (hcp, standard_hcp_scale, losers, controls, cccc, quality, exact_shape)
from endplay.dds import analyse_play

Expr = Callable[[Deal], Union[float, int, bool]]

class ConstraintInterpreter:
	"""
	Provides an interface for evaluating constraints in the Dealer program syntax
	and testing them against a deal
	"""

	# Precompiled regexes for matching some more complex function names
	_re_suit = re.compile(r"(?:spades?)|(?:hearts?)|(?:diamonds?)|(?:clubs)")
	_re_pt = re.compile(r"(?:pt[0-9])")
	_re_namedpt = re.compile(r"(?:tens?)|(?:jacks?)|(?:queens?)|(?:kings?)|(?:aces?)|(?:top[2-5])|(?:c13)")

	def __init__(self):
		self.parser = DealerParser()
		self.reset_env()

	def set_env(self, name: str, value: Any):
		"""
		Insert a new variable into the interpreter's environment or change the value 
		of an existing variable

		:param name: The name of the environment variable
		:param value: The new value of the variable
		"""
		self._env[name] = value

	def unset_env(self, name: str):
		"""
		Remove an variable from the interpreter's environment. Attempting to remove an
		essential variable (one of the ptN scale definitions) throws an error

		:param name: The name of the variable to remove
		"""
		if self._re_pt.match(name):
			raise RuntimeError(f"Trying to unset required environment variable {name}")
		del self._env[name]

	def get_env(self, name: str) -> Any:
		":return: The value of the named environment variable"
		return self._env[name]


	def reset_env(self):
		"Reinitialise the environment to the default values"
		self._env = {
			"pt0": [0,0,0,0,1,0,0,0,0,0,0,0,0], 
			"pt1": [0,0,0,1,0,0,0,0,0,0,0,0,0], 
			"pt2": [0,0,1,0,0,0,0,0,0,0,0,0,0],
			"pt3": [0,1,0,0,0,0,0,0,0,0,0,0,0], 
			"pt4": [1,0,0,0,0,0,0,0,0,0,0,0,0], 
			"pt5": [1,1,0,0,0,0,0,0,0,0,0,0,0],
			"pt6": [1,1,1,0,0,0,0,0,0,0,0,0,0], 
			"pt7": [1,1,1,1,0,0,0,0,0,0,0,0,0], 
			"pt8": [1,1,1,1,1,0,0,0,0,0,0,0,0],
			"pt9": [6,4,2,1,0,0,0,0,0,0,0,0,0]
		}

	def parse(self, s: str) -> Node:
		"Parse an expression string into a syntax tree"
		return self.parser.parse_expr(s)

	def evaluate(self, node: Union[Node, str], deal: Deal) -> float:
		"Evaluate an expression tree over a specific deal into a logical or arithmetic type"
		if isinstance(node, str):
			node = self.parse(node)
		if node.dtype == Node.ROOT:
			# root node is just dummy, keep going down the tree
			return self.evaluate(node.last_child)
		elif node.dtype == Node.SYMBOL:
			# symbols evaluate to their value in the environment and
			# are evaluated if they are Node instances
			val = self._env[node.value]
			if isinstance(val, Node):
				return self.evaluate(val, deal)
			else:
				return val
		elif node.dtype == Node.VALUE:
			# values evaluate to themselves
			return node.value
		elif node.dtype == Node.FUNCTION:
			# functions get dispatched
			return self._dispatch_function(node, deal)
		elif node.dtype == Node.OPERATOR:
			# operators get dispatched
			return self._dispatch_operator(node, deal)
		else:
			raise RuntimeError(f"Constraint contains unexpected node type {node.dtype}")

	def lambdify(self, node: Union[Node, str]) -> Expr:
		"""
		Convert an expression tree (or string) into an anonymous function accepting a single
		:class:`Deal` argument and returning the expression evaluated over this deal

		:param node: The root of the expression tree, or a string containing an expression
		"""
		if isinstance(node, str):
			node = self.parse(node)
		return lambda deal: self.evaluate(node, deal)

	def _evaluate_shape(self, node, shape):
		if node.dtype == Node.OPERATOR:
			if node.value == "any":
				tshape = shape.copy()
				pred = [x for x in node.first_child.value if x != None]
				n_wildcards = len(node.first_child.value) - len(pred)
				try:
					for elem in pred:
						tshape.remove(elem)
				except ValueError:
					return False
				return n_wildcards == len(tshape)
			else:
				lhs = self._evaluate_shape(node.first_child, shape)
				rhs = self._evaluate_shape(node.last_child, shape)
				if node.value == "+":
					return lhs or rhs
				else:
					return lhs and (not rhs)
				return rhs if node.value == "+" else (not rhs)
		elif node.dtype == Node.VALUE:
			for a, b in zip(shape, node.value):
				if b is not None and a != b:
					return False
			return True
		else:
			raise RuntimeError(f"Unexpected node type {node.dtype} in shape expression")


	def _dispatch_function(self, node, deal):
		if node.value == "hcp":
			return self._fn_hcp(node, deal, standard_hcp_scale)
		elif ConstraintInterpreter._re_suit.match(node.value):
			return self._fn_suit(node, deal)
		elif ConstraintInterpreter._re_pt.match(node.value):
			return self._fn_hcp(node, deal, self._env[node.value])
		elif ConstraintInterpreter._re_namedpt.match(node.value):
			if node.value == 'c13':
				idx = 9
			elif node.value[:-1] == 'top':
				idx = 3 + int(node.value[-1])
			else:
				idx = "tjqka".find(node.value[0])
			return self._fn_hcp(node, deal, self._env[f"pt{idx}"])
		elif node.value == "shape":
			return self._fn_shape(node, deal)
		elif node.value == "control" or node.value == "controls":
			return self._fn_control(node, deal)
		elif node.value == "loser" or node.value == "losers":
			return self._fn_loser(node, deal)
		elif node.value == "cccc":
			return self._fn_cccc(node, deal)
		elif node.value == "quality":
			return self._fn_quality(node, deal)
		elif node.value == "trick" or node.value == "tricks":
			return self._fn_trick(node, deal)
		elif node.value == "score":
			return self._fn_score(node, deal)
		elif node.value == "hascard":
			return self._fn_hascard(node, deal)
		elif node.value == "imp" or node.value == "imps":
			return self._fn_imp(node, deal)
		else:
			raise ValueError(f"Unknown function {node.value}")

	def _fn_hascard(self, node, deal):
		player = node.first_child.value
		card = node.last_child.value
		return card in deal[player]

	def _fn_hcp(self, node, deal, scale):
		if node.n_children == 1:
			return hcp(deal[node.first_child.value], scale)
		elif node.n_children == 2:
			return hcp(deal[node.first_child.value][node.last_child.value], scale)
		else:
			raise RuntimeError(f"hcp() given {node.n_children} arguments, expected 2")

	def _fn_suit(self, node, deal):
		suit = Denom.find(node.value)
		player = node.first_child.value
		return len(deal[player][suit])

	def _fn_control(self, node, deal):
		if node.n_children == 1:
			return controls(deal[node.first_child.value])
		elif node.n_children == 2:
			return controls(deal[node.first_child.value][node.last_child.value])
		else:
			raise RuntimeError(f"controls() given {node.n_children} arguments, expected 2")

	def _fn_loser(self, node, deal):
		if node.n_children == 1:
			return losers(deal[node.first_child.value])
		elif node.n_children == 2:
			return losers(deal[node.first_child.value][node.last_child.value])
		else:
			raise RuntimeError(f"loser() given {node.n_children} arguments, expected 2")

	def _fn_cccc(self, node, deal):
		hand = deal[node.first_child.value]
		return cccc(hand)
	
	def _fn_quality(self, node, deal):
		suit = deal[node.first_child.value][node.last_child.value]
		return quality(suit)

	def _fn_trick(self, node, deal):
		pos = node.first_child.value
		strain = node.last_child.value
		deal.first = pos.lho
		deal.trump = strain
		r = analyse_play(deal, [])
		return r[0]

	def _fn_score(self, node, deal):
		vul = node.first_child.value
		contract = node.middle_child.value
		result = node.last_child.value
		contract.result = result
		return contract.score(vul)

	def _fn_imps(self, node, deal):
		raise NotImplementedError

	def _fn_shape(self, node, deal):
		hand = deal[node.first_child.value]
		s = exact_shape(hand)
		return self._evaluate_shape(node.last_child, s)

	def _fn_if(self, node, deal):
		if self.evaluate(node.first_child):
			return self.evaluate(node.middle_child, deal)
		else:
			return self.evaluate(node.last_child, deal)

	def _dispatch_operator(self, node, deal):
		if node.value in ["&&", "and"]:
			return self._op_and(node, deal)
		elif node.value in ["||", "or"]:
			return self._op_or(node, deal)
		elif node.value in ["!", "not"]:
			return self._op_not(node, deal)
		elif node.value == "==":
			return self._op_equal(node, deal)
		elif node.value == "!=":
			return self._op_notequal(node, deal)
		elif node.value == "<":
			return self._op_less(node, deal)
		elif node.value == "<=":
			return self._op_leq(node, deal)
		elif node.value == ">":
			return self._op_greater(node, deal)
		elif node.value == ">=":
			return self._op_geq(node, deal)
		elif node.value == "+":
			return self._op_add(node, deal)
		elif node.value == "-":
			return self._op_sub(node, deal)
		elif node.value == "*":
			return self._op_mul(node, deal)
		elif node.value == "/":
			return self._op_div(node, deal)
		elif node.value == "%":
			return self._op_mod(node, deal)
		else:
			raise ValueError(f"Unknown operator {node.value}")

	def _op_and(self, node, deal):
		return self.evaluate(node.first_child, deal) and self.evaluate(node.last_child, deal)
		
	def _op_or(self, node, deal):
		return self.evaluate(node.first_child, deal) or self.evaluate(node.last_child, deal)

	def _op_not(self, node, deal):
		return not self.evaluate(node.first_child, deal)

	def _op_equal(self, node, deal):
		return self.evaluate(node.first_child, deal) == self.evaluate(node.last_child, deal)

	def _op_notequal(self, node, deal):
		return self.evaluate(node.first_child, deal) != self.evaluate(node.last_child, deal)

	def _op_less(self, node, deal):
		return self.evaluate(node.first_child, deal) < self.evaluate(node.last_child, deal)

	def _op_leq(self, node, deal):
		return self.evaluate(node.first_child, deal) <= self.evaluate(node.last_child, deal)

	def _op_greater(self, node, deal):
		return self.evaluate(node.first_child, deal) > self.evaluate(node.last_child, deal)
	
	def _op_geq(self, node, deal):
		return self.evaluate(node.first_child, deal) >= self.evaluate(node.last_child, deal)

	def _op_add(self, node, deal):
		return self.evaluate(node.first_child, deal) + self.evaluate(node.last_child, deal)

	def _op_sub(self, node, deal):
		return self.evaluate(node.first_child, deal) - self.evaluate(node.last_child, deal)

	def _op_mul(self, node, deal):
		return self.evaluate(node.first_child, deal) * self.evaluate(node.last_child, deal)

	def _op_div(self, node, deal):
		return self.evaluate(node.first_child, deal) / self.evaluate(node.last_child, deal)

	def _op_mod(self, node, deal):
		return self.evaluate(node.first_child, deal) % self.evaluate(node.last_child, deal)



__all__ = ["Penalty"]

from enum import IntEnum

class Penalty(IntEnum):
	"Encodes a penalty"
	passed = 1
	doubled = 2
	redoubled = 4

	@staticmethod
	def find(name: str):
		if not name or name[0].lower() in "np" or name.lower() == "ap":
			return Penalty.passed
		if name.lower() == 'x' or name[0].lower() == 'd':
			return Penalty.doubled
		if name.lower() == 'xx' or name[0].lower() == 'r':
			return Penalty.redoubled
		raise ValueError(f"Invalid Penalty: {name}")

	@property
	def abbr(self) -> str:
		return "x" * (self >> 1)



from __future__ import annotations

__all__ = ["Deal"]

import ctypes
import json as _json
import sys
from collections.abc import Iterable, Iterator
from typing import Optional, Union

import endplay._dds as _dds
from endplay.types.card import Card
from endplay.types.denom import Denom
from endplay.types.hand import Hand
from endplay.types.player import Player
from endplay.types.rank import AlternateRank
from endplay.types.vul import Vul


class Deal:
	"""
	Class representing a bridge deal. The class keeps track of the four hands
	at the table, as well as

	* The player on initial lead to the current trick in `Deal.first`
	* The trump suit the deal is played in in `Deal.trump`
	* The cards played to the current trick in `Deal.curtrick`
	"""
	def __init__(self,
		pbn: Optional[str] = None,
		first: Player = Player.north,
		trump: Denom = Denom.nt,
		*,
		complete_deal: bool = False):
		self._data = _dds.deal()
		self.clear()
		self._data.first = first
		self._data.trump = trump
		if pbn is not None:
			if len(pbn) > 2 and pbn[1] == ":":
				start, hands = Player.find(pbn[0]), pbn[2:].split()
			else:
				start, hands = Player.north, pbn.split()
			if len(hands) > 4:
				raise RuntimeError(f"deal contains more than four hands: {pbn}")
			for player, hand in zip(Player.iter_from(start), hands):
				if hand != "-":
					self[player] = hand
			if complete_deal:
				self.complete_deal()

	def __copy__(self) -> 'Deal':
		other = Deal()
		other._data.trump = self._data.trump
		other._data.first = self._data.first
		size_of_trick = ctypes.sizeof(ctypes.c_int) * 3
		size_of_cards = ctypes.sizeof(ctypes.c_uint) * 4 * 4
		ctypes.memmove(other._data.currentTrickSuit, self._data.currentTrickSuit, size_of_trick)
		ctypes.memmove(other._data.currentTrickRank, self._data.currentTrickRank, size_of_cards)
		ctypes.memmove(other._data.remainCards, self._data.remainCards, size_of_cards)
		return other

	def copy(self) -> 'Deal':
		"Return a new copy of this deal"
		return self.__copy__()

	@property
	def trump(self) -> Denom:
		"The trump suit the deal is being played in"
		return Denom(self._data.trump)
	@trump.setter
	def trump(self, denom: Denom) -> None:
		self._data.trump = denom

	@property
	def first(self) -> Player:
		"The player to lead the first card to the current trick"
		return Player(self._data.first)
	@first.setter
	def first(self, player) -> None:
		self._data.first = player

	@property
	def curplayer(self) -> Player:
		"The player to play the next card to the current trick"
		return Player(self._data.first).next(len(self.curtrick))

	@property
	def curhand(self) -> Hand:
		"The hand of the `curplayer`"
		return self[self.curplayer]

	@property
	def north(self) -> Hand:
		"The north hand"
		return self[Player.north]
	@north.setter
	def north(self, hand: Hand) -> None:
		self[Player.north] = hand

	@property
	def east(self) -> Hand:
		"The east hand"
		return self[Player.east]
	@east.setter
	def east(self, hand: Hand) -> None:
		self[Player.east] = hand

	@property
	def south(self) -> Hand:
		"The south hand"
		return self[Player.south]
	@south.setter
	def south(self, hand: Hand) -> None:
		self[Player.south] = hand

	@property
	def west(self) -> Hand:
		"The west hand"
		return self[Player.west]
	@west.setter
	def west(self, hand: Hand) -> None:
		self[Player.west] = hand

	@property
	def curtrick(self) -> list[Card]:
		"Return a list of cards played to the current trick"
		trick = []
		for i in range(3):
			suit, rank = self._data.currentTrickSuit[i], self._data.currentTrickRank[i]
			if rank == 0:
				break
			trick.append(Card(suit=Denom(suit), rank=AlternateRank(rank).to_standard()))
		return trick

	def play(self, card: Card, fromHand: bool = True) -> None:
		"""
		Play a card to the current trick. If the played card completes the trick then the entire
		trick is cleared and `first` is set to the played who won the trick using `trump` as the
		trumps suit

		:param card: The cad to be played
		:param fromHand: If true then the card will be removed from the hand of the player who is to
			play to the trick next and throw a RuntimeError if they do not hold this card.
		"""
		if isinstance(card, str):
			card = Card(card)
		for i in range(3):
			if self._data.currentTrickRank[i] == 0:
				if fromHand and not self[self.first.next(i)].remove(card):
					raise RuntimeError("Trying to play card not in hand")
				self._data.currentTrickRank[i] = card.rank.to_alternate()
				self._data.currentTrickSuit[i] = card.suit
				break
		else:
			# Playing final card to a trick: Calculate winner, update first and clear current trick
			# Lazy import to avoid circular import
			from endplay.utils.play import trick_winner
			if fromHand and not self[self.first.rho].remove(card):
				raise RuntimeError("Trying to play card not in hand")
			self.first = trick_winner(self.curtrick + [card], self.first, self.trump)
			for i in range(3):
				self._data.currentTrickRank[i] = 0

	def unplay(self, toHand: bool = True) -> Card:
		"""
		Unplay the last card played to the current trick. Throws a RuntimeError if the current
		trick is empty

		:param toHand: If true then the card is returned to the hand of the player who played
			it to the trick
		:return: The card that was picked up
		"""
		for i in reversed(range(3)):
			suit, rank = self._data.currentTrickSuit[i], self._data.currentTrickRank[i]
			if rank != 0:
				self._data.currentTrickRank[i] = 0
				card = Card(suit=Denom(suit), rank=AlternateRank(rank).to_standard())
				if toHand:
					self[self.first.next(i)].add(card)
				return card
		else:
			raise RuntimeError("No cards to unplay")

	def complete_deal(self) -> None:
		"""
		If there is a player with no cards, deal any cards which do not appear in anyone else's
		hand to that player
		"""
		remaining = Hand("AKQJT98765432.AKQJT98765432.AKQJT98765432.AKQJT98765432")
		missing_hand = None
		for player, hand in self:
			if len(hand) == 0:
				if missing_hand:
					raise ValueError("Cannot complete a deal with more than one missing hand")
				missing_hand = player
			else:
				for card in hand:
					remaining.remove(card)
		if missing_hand is not None:
			if len(remaining) != 13:
				raise ValueError("Cannot complete a deal with more than 13 undealt cards")
			self[missing_hand] = remaining

	@staticmethod
	def from_pbn(pbn: str) -> 'Deal':
		"""
		Clear all hands and the current trick, and replace with the cards in a PBN string

		:param pbn: A PBN string, e.g. "N:974.AJ3.63.AK963 K83.K9752.7.8752 AQJ5.T864.KJ94.4 T62.Q.AQT852.QJT"
		"""
		return Deal(pbn)

	def to_pbn(self) -> str:
		"Return a PBN string representation of the deal"
		return 'N:' + ' '.join(str(self[player]) for player in Player)

	@staticmethod
	def from_json(json: Union[str, dict]):
		if isinstance(json, str):
			json = _json.loads(json)
		deal = Deal()
		deal.north = json["north"]
		deal.east = json["east"]
		deal.south = json["south"]
		deal.west = json["west"]
		deal.first = Player.find(json["first"])
		deal.trump = Denom.find(json["trump"])
		for card in json["curtrick"]:
			deal.play(card, False)
		return deal

	def to_json(self, indent: Optional[int] = None) -> str:
		"Encode the data as a JSON string"
		d = {
			"north": self.north.to_pbn(),
			"east": self.east.to_pbn(),
			"south": self.south.to_pbn(),
			"west": self.west.to_pbn(),
			"curtrick": [str(card) for card in self.curtrick],
			"first": self.first.name,
			"trump": self.trump.name
		}
		return _json.dumps(d, indent=indent)

	@staticmethod
	def from_lin(lin: str, complete_deal: bool = True):
		"""
		Construct a deal from a LIN format deal string.

		:param complete_deal: If True, then add remaining cards to the last hand if it is
			omitted from the input string (which is done by default for files downloaded
			from BBO)
		"""
		if lin[0] in "1234":
			hands = lin[1:].split(",")
		else:
			hands = lin.split(",")
		deal = Deal()
		for player, hand in zip(Player.iter_from(Player.south), hands):
			deal[player] = Hand.from_lin(hand)
		if complete_deal:
			deal.complete_deal()
		return deal

	def to_lin(self, dealer: Optional[Player] = None, complete_deal: bool = False) -> str:
		"""
		Convert a deal to LIN representation as used by BBO

		:param dealer: If provided, append the dealer (in LIN format) to the returned string
		:param complete_deal: If False, omit the last hand from the string. This is the default
			for files created by BBO.
		"""
		lin = "" if dealer is None else str(dealer.to_lin())
		lin += self[Player.south].to_lin() + ","
		lin += self[Player.west].to_lin() + ","
		lin += self[Player.north].to_lin() + ","
		if complete_deal:
			lin += self[Player.east].to_lin()
		return lin

	def to_LaTeX(self, board_no: Optional[int] = None, exclude: Iterable[Player] = [], ddtable: bool = False) -> str:
		"Return a LaTeX representation of the hand"
		mkhand = lambda p: self[p].to_LaTeX() if p not in exclude else ""
		res = r"\begin{tabular}{c c c}"
		if board_no is not None:
			res += r"\begin{tabular}{c}"
			res += f"Board {board_no}\\\\Vul "
			vul = Vul.from_board(board_no)
			if vul == Vul.none: res += "none"
			elif vul == Vul.both: res += "all"
			elif vul == Vul.ns: res += "N/S"
			else: res += "E/W"
			res += f"\\\\{Player.west.next(board_no).abbr} deals\\\\"
			res += r"\end{tabular}"
		res += " & " + mkhand(Player.north) + " & \\\\"
		res += mkhand(Player.west) + r" & \fbox{\begin{tabular}{c c c}&N&\\W&&E\\&S&\end{tabular}} & "
		res += mkhand(Player.east) + r"\\ & " + mkhand(Player.south) + " & "
		if ddtable:
			from endplay.dds.ddtable import calc_dd_table
			res += r"\footnotesize{\setlength{\tabcolsep}{2pt} " + calc_dd_table(self).to_LaTeX() + "}"
		res += r"\end{tabular}"
		return res

	def to_hand(self) -> Hand:
		"Return a new hand containing the contents of all four hands in the deal"
		res = Hand()
		for _, hand in self:
			res.extend(hand)
		return res

	def rotate(self, n: int) -> None:
		"Rotate clockwise by n quarter-turns, e.g. with n=1 NESW -> WNES"
		n %= 4
		if n == 0:
			return
		rc = self._data.remainCards
		# tuple assignment doesn't work for ctypes types so we use tmp + memmove
		# also as the array is only 4 elements it's not too much effort to treat
		# the four cases separately for a bit of extra efficiency
		l = ctypes.sizeof(ctypes.c_uint) * 4
		tmp = (ctypes.c_uint * 4).from_buffer_copy(rc[0])
		if n == 1: # nesw -> wnes
			ctypes.memmove(rc[0], rc[3], l)
			ctypes.memmove(rc[3], rc[2], l)
			ctypes.memmove(rc[2], rc[1], l)
			ctypes.memmove(rc[1], tmp, l)
		elif n == 2: # nesw -> swne
			self.swap(Player.north, Player.south)
			self.swap(Player.east, Player.west)
		elif n == 3: # nesw -> eswn:
			ctypes.memmove(rc[0], rc[1], l)
			ctypes.memmove(rc[1], rc[2], l)
			ctypes.memmove(rc[2], rc[3], l)
			ctypes.memmove(rc[3], tmp, l)

	def swap(self, player_a: Player, player_b: Player) -> None:
		"Swap the specified hands"
		rc = self._data.remainCards
		# tuple assignment doesn't work for ctypes types so we use tmp + memmove
		l = ctypes.sizeof(ctypes.c_uint) * 4
		tmp = (ctypes.c_uint * 4).from_buffer_copy(rc[player_a])
		ctypes.memmove(rc[player_a], rc[player_b], l)
		ctypes.memmove(rc[player_b], tmp, l)

	def clear(self) -> None:
		"Clear all hands and the cards in the current trick"
		for _, hand in self:
			hand.clear()
		while self.curtrick:
			self.unplay(False)

	def pprint(self,
		board_no: Optional[int] = None,
		exclude: list[Player] = [],
		stream=sys.stdout,
		*,
		vul: Optional[Vul] = None,
		dealer: Optional[Player] = None) -> None:
		"""
		Print the deal in a hand diagram format.

		:param board_no: If provided, the hand diagram will display the board
			number, vulnerability and dealer in the top-left corner
		"""
		spacing = " " * 13
		played_cards = ["  "] * 4
		prefixes = ["	^", "	   >", "	v", "  <"]
		titles = [ spacing ] * 4
		if board_no is not None:
			titles[0] = f"Board {board_no}".center(13)
			if vul is None: vul = Vul.from_board(board_no)
			if dealer is None: dealer = Player.west.next(board_no)
		if vul is not None:
			titles[1] = f"Vul {vul.abbr}".center(13)
		if dealer is not None:
			titles[2] = f"{dealer.abbr} deals".center(13)
		for player, card in zip(Player.iter_from(self.first), self.curtrick):
			played_cards[player] = prefixes[player] + str(card)
		for title, suit in zip(titles, Denom.suits()):
			n = (str(self[Player.north][suit]) or "---") if Player.north not in exclude else ""
			print(title, n, file=stream)
		for player, suit in zip([Player.find(p) for p in "news"], Denom.suits()):
			w = (str(self[Player.west][suit]) or "---") if Player.west not in exclude else ""
			e = (str(self[Player.east][suit]) or "---") if Player.east not in exclude else ""
			print(w.ljust(13), played_cards[player].ljust(13), e, file=stream)
		for suit in Denom.suits():
			s = (str(self[Player.south][suit]) or "---") if Player.south not in exclude else ""
			print(spacing, s, file=stream)

	def __contains__(self, card: Card) -> bool:
		":return: True if card is in the current deal"
		return any(card in hand for _, hand in self)

	def __iter__(self) -> Iterator[tuple[Player, Hand]]:
		":return: An iterator over the north, east, south and west hands respectively"
		yield from ((player, self[player]) for player in Player)

	def __getitem__(self, player: Player) -> Hand:
		":return: The specified hand"
		return Hand(self._data.remainCards[player])

	def __setitem__(self, player: Player, hand: Union[Hand, str]) -> None:
		"Set the hand to the specified hand, which may be in the format of a PBN string"
		if isinstance(hand, str):
			hand = Hand(hand)
		self._data.remainCards[player] = hand._data

	def compare(self, other: Deal, hands_only: bool = False) -> bool:
		cmp = _dds._libc.memcmp
		if cmp(self._data.remainCards, other._data.remainCards, len(self._data.remainCards)):
			return False
		if not hands_only:
			if self._data.trump != other._data.trump:
				return False
			if self._data.first != other._data.first:
				return False
			if cmp(self._data.currentTrickSuit, other._data.currentTrickSuit, len(self._data.currentTrickSuit)):
				return False
			if cmp(self._data.currentTrickRank, other._data.currentTrickRank, len(self._data.currentTrickRank)):
				return False
		return True

	def __eq__(self, other: Deal) -> bool:
		return self.compare(other)

	def __repr__(self) -> str:
		return f"Deal('{self!s}')"

	def __str__(self) -> str:
		":return: A PBN string representation of the deal"
		return self.to_pbn()



"""
Double dummy table calculation functions
"""

from __future__ import annotations

__all__ = ["DDTable", "DDTableList", "calc_dd_table", "calc_all_tables"]

import sys
from collections import abc
from collections.abc import Iterable
import endplay._dds as _dds
from endplay.types import Deal, Denom, Player

class DDTable:
	"""
	Python wrapper for the `_dds.ddTableResults` class. Entries can be accessed
	using the `__getitem__` operator e.g. table[Denom.clubs, Player.west]
	"""
	def __init__(self, data: _dds.ddTableResults):
		self._data = data
		
	def pprint(self, 
		*, 
		denoms: Iterable[Denom] = [Denom.clubs, Denom.diamonds, Denom.hearts, Denom.spades, Denom.nt],
		players: Iterable[Player] = [Player.north, Player.south, Player.east, Player.west],
		stream=sys.stdout) -> None:
		"""
		Print the double dummy table in a grid format
		
		:param denoms: Specify the columns of the table
		:param players: Specify the rows of the table
		"""
		denoms, players = list(denoms), list(players)
		print("   ", " ".join(denom.abbr.rjust(2) for denom in denoms), file=stream)
		for player in players:
			print(player.abbr.rjust(3), end='', file=stream)
			for denom in denoms:
				print(str(self[denom, player]).rjust(3), end='', file=stream)
			print(file=stream)
	
	def to_LaTeX(self) -> str:
		"""Create a LaTeX string of the table"""
		res = r"\begin{tabular}{| c | c  c  c  c c |}"
		res += r"\hline & $\clubsuit$ & $\diamondsuit$ & $\heartsuit$ & $\spadesuit$ & NT \\ \hline "
		for player in Player.iter_order("NSEW"):
			res += player.abbr + " & " + " & ".join(str(self[denom][player]) for denom in Denom.bidorder()) + "\\\\"
		res += r"\hline \end{tabular}"
		return res

	def to_list(self, player_major: bool = False) -> list[list[int]]:
		"""
		Convert the table to a 2d list

		:param player_major: If `True`, the returned list is index by player first then strain
		"""
		if player_major:
			return [[self[d, p] for d in Denom] for p in Player]
		else:
			return [[self[d, p] for p in Player] for d in Denom]

	def __getitem__(self, cell: tuple[Denom, Player]) -> int:
		"""Return the specified cell of the table"""
		if isinstance(cell[0], Denom):
			return self._data.resTable[cell[0]][cell[1]]
		else:
			return self._data.resTable[cell[1]][cell[0]]

	def __str__(self) -> str:
		return ",".join(d.abbr for d in Denom.bidorder()) + \
			";" + ";".join(p.abbr + ":" + \
			",".join(str(self[d, p]) for d in Denom.bidorder()) for p in Player)

class DDTableList(abc.Sequence):
	def __init__(self, data: '_dds.ddTableRes'):
		self._data = data

	def __len__(self) -> int:
		"The number of double dummy tables in the list"
		return self._data.noOfBoards

	def __getitem__(self, i: int) -> DDTable:
		"Return the double dummy table at index `i`"
		if i < 0:
			i = len(self) - i
		if i < 0 or i >= len(self):
			raise IndexError
		return DDTable(self._data.results[i])

	def __repr__(self) -> str:
		return f'<DDTableList object; length={len(self)}>'

	def __str__(self) -> str:
		if len(self) == 0:
			return "[]"
		return '[(' + "), (".join(str(t) for t in self) + ")]"

def calc_dd_table(deal: Deal) -> DDTable:
	"""
	Calculates the double dummy results for all 20 possible combinations of
	dealer and trump suit for a given deal
	"""
	# Convert deal into ddTableDeal
	dl = _dds.ddTableDeal()
	dl.cards = deal._data.remainCards

	table =_dds.ddTableResults()
	_dds.CalcDDtable(dl, table)
	return DDTable(table)

def calc_all_tables(deals: Iterable[Deal], exclude: Iterable[Denom] = []) -> DDTableList:
	"""
	Optimized version of calc_dd_table for multiple deals which uses threading to
	speed up the calculation. `exclude` can contain a list of denominations to
	exclude from the calculation, e.g. if only the notrump results for the deals
	is required then pass `Denom.suits()` 
	"""
	# Convert deals to ddTableDeals
	dealsp = _dds.ddTableDeals()
	dealsp.noOfTables = 0
	for i, deal in enumerate(deals):
		if i > _dds.MAXNOOFTABLES * 5:
			raise RuntimeError(f"Too many boards, maximum is {_dds.MAXNOOFTABLES * 5}")
		dealsp.ddTableDeal[i].cards = deal._data.remainCards
		dealsp.noOfTables += 1

	# Convert exclude to a trump filter list
	trumpFilter = [False]*5
	for trump in exclude:
		trumpFilter[trump] = True

	resp = _dds.ddTablesRes()
	presp = _dds.allParResults()
	_dds.CalcAllTables(dealsp, -1, trumpFilter, resp, presp)
	resp.noOfBoards = dealsp.noOfTables
	return DDTableList(resp)


"""
Algorithms for detecting play techniques
"""

from __future__ import annotations

__all__ = ["detect_play", "SimpleFinesse", "PlayTechniqueBase"]

from endplay.types import Deal, Card, Rank, SuitHolding

class NoMatch(Exception):
	pass

def rank_below(rank: Rank, holding: SuitHolding):
	"Return the rank just below `rank` in `holding`. Returns `None` if `rank` is the first card"
	prev = None
	for r in holding:
		if r == rank:
			return prev
		prev = r
	raise ValueError(f"{rank} not found in holding")

def rank_above(rank: Rank, holding: SuitHolding):
	"Return the rank just above `rank` in `holding`. Returns `None` if `rank` is the last card"
	foundRank = None
	for r in holding:
		if r == rank:
			foundRank = True
		elif foundRank:
			return r
	if foundRank:
		return None
	raise ValueError(f"{rank} not found in holding")

def rank_equivalent(rankA: Rank, rankB: Rank, remaining: SuitHolding):
	"""
	Return true if `cardA` and `cardB` are equivalent cards, given that
	the only cards of that suit left in the deal are in `remaining`
	"""
	foundA = False
	for rank in remaining:
		if rank == rankA:
			foundA = True
		elif foundA:
			return rank == rankB
	raise ValueError(f"{rankA} not found in remaining cards")


class PlayTechniqueBase:
	def __init__(self):
		self.match = False
		self.exception = None

	def __bool__(self):
		return self.match

	def match(self, deal, trick):
		try:
			if len(trick) != 4:
				raise ValueError("`trick` must contain four cards")
			for i, card in enumerate(trick):
				if not isinstance(card, Card):
					trick[i] = Card(card)
			self._match(deal, trick)
		except NoMatch as e:
			self.exception = e
		return self.match


class Finesse(PlayTechniqueBase):
	def __init__(self):
		super().__init__()
		self.onside = False
		self.lho_covered = False
		self.rho_covered = False

class SimpleFinesse(PlayTechniqueBase):
	def __init__(self):
		super().__init__()
		self.is_onside = False
		self.lho_covered = False
		self.rho_covered = False

	def _match(self, deal: Deal, trick: list[Card]):
		# First three cards must be the same suit
		suit = trick[0].suit
		if suit != trick[1].suit or suit != trick[2].suit:
			raise NoMatch("First three cards must be of the same suit")
		# Get all cards in the suit
		lho = deal[deal.first.lho][suit]
		rho = deal[deal.first.rho][suit]
		opps = lho.copy()
		opps.extend(rho)
		leader = deal[deal.first][suit]

		# Check for the following scenarios:
		# - 

		leader.extend(deal[deal.first.partner][suit])
		total = opps.copy()
		total.extend(leader)
		# Find cards in the opponent's hands where the leader has the
		# card directly above/below


def detect_play(deal: Deal, play_history: list[Card], techniques: list[PlayTechniqueBase]=None):
	if techniques is None:
		techniques = [SimpleFinesse]

	res = []
	for technique in techniques:
		match = technique()
		if technique.match(deal, play_history):
			res.append(match)
	return res


from __future__ import annotations

__all__ = ["Contract"]

import re
from typing import Union, Optional
from collections.abc import Sequence

from endplay.types.player import Player
from endplay.types.denom import Denom
from endplay.types.penalty import Penalty
from endplay.types.vul import Vul
from endplay.types.bid import Bid, ContractBid
import endplay._dds as _dds

contract_to_denom = [ Denom.nt, Denom.spades, Denom.hearts, Denom.diamonds, Denom.clubs ]
denom_to_contract = [ 1, 2,3,4,0 ]

class Contract:
	"Class representing a specific contract"
	_pat = re.compile(r"^([1-7])((?:NT?)|S|H|D|C)([NSEW]?)((?:XX|X|D|R)?)((?:=|(?:[+-]\d+))?)$")
	def __init__(
		self,
		data: Union[_dds.contractType, str, None] = None,
		*,
		level: Optional[int] = None,
		denom: Optional[Denom] = None,
		declarer: Optional[Player] = None,
		penalty: Optional[Penalty] = None,
		result: Optional[int] = None):
		"""
		Construct a new contract. If no parameters are passed, a passout is constructed.

		:param data: Construct from a _dds.contractType object or contract string
		:param level: The level of the contract
		:param denom: The denomination of the contract
		:param declarer: The declarer of the contract
		:param result: The number of overtricks (positive) or undertricks (negative) made
		:param penalty: The penalty (pass, x or xx) of the contract
		"""
		if isinstance(data, str):
			m = Contract._pat.match(data.upper())
			if m:
				level = int(m[1])
				denom = Denom.find(m[2])
				declarer = Player.find(m[3] or "north")
				penalty = Penalty.find(m[4])
				result = 0 if m[5] == "=" else int(m[5] or "0")
			elif data.lower().startswith("pass"):
				level = 0
			else:
				raise ValueError(f"Invalid contract string: {data}")

			data = None

		self._data = data or _dds.contractType()
		self.penalty = Penalty.passed if self._data.underTricks == 0 else Penalty.doubled
		if level is not None: self.level = level
		if denom is not None: self.denom = denom
		if declarer is not None: self.declarer = declarer
		if penalty is not None: self.penalty = penalty
		if result is not None: self.result = result

	def copy(self) -> Contract:
		"Return a copy of this contract object"
		return self.__copy__()

	def __copy__(self) -> Contract:
		_data = _dds.contractType()
		_data.underTricks = self._data.underTricks
		_data.overTricks = self._data.overTricks
		_data.level = self._data.level
		_data.denom = self._data.denom
		_data.seats = self._data.seats
		return Contract(_data)

	@property
	def level(self) -> int:
		"The level of the contract"
		return self._data.level
	@level.setter
	def level(self, new_level: int) -> None:
		self._data.level = new_level

	@property
	def denom(self) -> Denom:
		"The denomination of the contract"
		return contract_to_denom[self._data.denom]
	@denom.setter
	def denom(self, new_denom: Denom) -> None:
		self._data.denom = denom_to_contract[new_denom]

	@property
	def declarer(self) -> Player:
		"The declarer of the contract"
		return Player(self._data.seats)
	@declarer.setter
	def declarer(self, player: Player):
		self._data.seats = player

	@property
	def result(self) -> int:
		"The number of tricks over or under the contract made"
		if self._data.underTricks != 0:
			return -self._data.underTricks
		if self._data.overTricks != 0:
			return self._data.overTricks
		return 0
	@result.setter
	def result(self, tricks: int):
		if tricks > 0:
			self._data.underTricks = 0
			self._data.overTricks = tricks
		else:
			self._data.underTricks = -tricks
			self._data.overTricks = 0

	@staticmethod
	def from_auction(dealer: Player, auction: Sequence[Bid]):
		"""
		Construct a contract from a bidding sequence
		"""
		# Iterate backwards through the sequence. The penalty is the
		# first x/xx bid we find, and the contract is the first
		# contract bid we find
		c = Contract()
		last = dealer.next(len(auction)-1)
		for player, bid in last.enumerate(reversed(auction), step=-1):
			if isinstance(bid, ContractBid):
				c.denom = bid.denom
				c.level = bid.level
				c.declarer = player
				break
			elif c.penalty == Penalty.passed:
				c.penalty = bid.penalty
		# Iterate forwards through the sequence. If the denom is the same
		# as the contract and the bidder is the (current) declarer or their
		# partner, set the declarer to that player
		for player, bid in dealer.enumerate(auction):
			if isinstance(bid, ContractBid) and bid.denom == c.denom and player in [c.declarer, c.declarer.partner]:
				c.declarer = player
				break
		return c

	def is_passout(self) -> bool:
		"Returns true if the contract represents a passout"
		return self.level == 0

	def score(self, vul: Vul) -> int:
		"The number of points the contract would score for the declarer"
		if self.is_passout():
			return 0
		if vul == Vul.none:
			is_vul = False
		elif vul == Vul.ns:
			is_vul = (self.declarer in (Player.north, Player.south))
		elif vul == Vul.ew:
			is_vul = (self.declarer in (Player.east, Player.west))
		else:
			is_vul = True
		res = self.result
		if res < 0:
			if self.penalty == Penalty.passed:
				if is_vul:
					return 100 * res
				else:
					return 50 * res
			elif self.penalty == Penalty.doubled:
				if is_vul:
					s = [-200] + [-300] * 12
					return sum(s[:-res])
				else:
					s = [-100] + [-200] * 2 + [-300] * 10
					return sum(s[:-res])
			elif self.penalty == Penalty.redoubled:
				if is_vul:
					s = [-400] + [-600] * 12
					return sum(s[:-res])
				else:
					s = [-200] + [-400] * 2 + [-600] * 10
					return sum(s[:-res])
			else:
				raise ValueError(f"Unknown penalty {self.penalty}")
		else:
			d, l = self.denom, self.level
			# Contract score
			if d == Denom.clubs or d == Denom.diamonds:
				score = 20 * l
			else:
				score = 30 * l
			if d == Denom.nt:
				score += 10
			score *= self.penalty
			# Game/part-score bonus
			if score >= 100:
				if is_vul: score += 500
				else: score += 300
			else:
				score += 50
			# Slam bonuses
			if l == 6:
				if is_vul: score += 750
				else: score += 500
			elif l == 7:
				if is_vul: score += 1500
				else: score += 750
			# Insult bonus
			if self.penalty == Penalty.doubled:
				score += 50
			elif self.penalty == Penalty.redoubled:
				score += 100
			# Overtrick bonus
			if self.penalty == Penalty.passed:
				if self.denom == Denom.clubs or self.denom == Denom.diamonds:
					score += 20 * res
				else:
					score += 30 * res
			elif self.penalty == Penalty.doubled:
				if is_vul: score += 200 * res
				else: score += 100 * res
			else:
				if is_vul: score += 400 * res
				else: score += 200 * res
			return score

	def __eq__(self, other: Contract) -> bool:
		return \
			self.level == other.level and \
			self.denom == other.denom and \
			self.declarer == other.declarer and \
			self.result == other.result

	def __repr__(self) -> str:
		return f'Contract("{self!s}")'

	def __str__(self) -> str:
		"A string representation of the contract e.g. 3NTW="
		if self.is_passout():
			return "Pass"
		s = f"{self.level}{self.denom.abbr}{self.declarer.abbr}{self.penalty.abbr}"
		return s + ("=" if self.result == 0 else f"{self.result:+d}")



"""
Actions class for producing plaintext output.
"""

__all__ = ["TerminalActions"]

from io import StringIO
from endplay.dealer.actions.base import BaseActions
from endplay.types import Player
import endplay.stats as stats

class TerminalActions(BaseActions):
	def __init__(self, deals, stream, board_numbers):
		super().__init__(deals, stream, board_numbers)

	def print(self, *players):
		exclude = [p for p in Player if p not in players]
		for i, deal in enumerate(self.deals, 1):
			deal.pprint(board_no=i if self.board_numbers else None, exclude=exclude, stream=self.stream)
			self.write()

	def printpbn(self):
		for i, deal in enumerate(self.deals, 1):
			if self.board_numbers:
				self.write(str(i).rjust(3), end=' ')
			self.write(str(deal))

	def printcompact(self, expr = None):
		self.write(*[p.name.ljust(13) for p in Player], end="\n\n")
		for i, deal in enumerate(self.deals, 1):
			hands = []
			for player in Player:
				buf = StringIO()
				deal[player].pprint(stream=buf)
				hands.append(buf.getvalue().split("\n"))
			if expr is not None:
				hands.append([str(expr(deal)), "", "", ""])
			if self.board_numbers:
				self.write(str(i) + ")")
			for line in zip(*hands):
				self.write(*[s.ljust(13) for s in line])
			self.write()

	def printoneline(self, expr = None):
		for i, deal in enumerate(self.deals, 1):
			if self.board_numbers:
				self.write(str(i).rjust(3), end=' ')
			self.write(deal.to_pbn()[2:], end=' ')
			if expr is not None:
				self.write(expr(deal))
			else:
				self.write()

	def printes(self, *objs):
		for i, deal in enumerate(self.deals, 1):
			if self.board_numbers:
				self.write(str(i).rjust(3), end=' ')
			for obj in objs:
				if callable(obj):
					self.write(obj(deal), end='')
				else:
					self.write(obj, end='')
			self.write()

	def average(self, expr, s = None):
		if s:
			self.write(s, end="")
		self.write(stats.average(self.deals, expr))

	def frequency1d(self, expr, lb, ub, s = None):
		hist = stats.frequency(self.deals, expr, lb, ub)
		if s:
			self.write(s)
			self.write("=" * len(s))
		rows = [(str(start), str(val)) for start, val in enumerate(hist, lb)]
		lhs_size = max(len(row[0]) for row in rows)
		for row in rows:
			self.write(row[0].rjust(lhs_size), row[1])

	def frequency2d(self, ex1, lb1, ub1, ex2, lb2, ub2, s = None):
		hist = stats.cofrequency(self.deals, ex1, ex2, lb1, ub1, lb2, ub2)
		if s:
			self.write(s)
			self.write("=" * len(s))
		rows = [[""] + [str(i) for i in range(lb2, ub2+1)]]
		for j, row in enumerate(hist, lb1):
			rows += [[str(j) + " |"] + [str(r) for r in row]]
		width = max(max(len(cell) for cell in row) for row in rows)
		self.write(" ".join(c.rjust(width) for c in rows[0]))
		self.write("+".rjust(width) + "-" + "-".join("-"*width for _ in rows[0][1:]))
		for row in rows[1:]:
			self.write(" ".join(c.rjust(width) for c in row))




"""
All-purpose routine for executing a dealer script file. This can be
called directly from within Python to convert a dealer file into a 
list of deals, but its main purpose is as the entry point for the
main module.
"""

from __future__ import annotations

from typing import Optional
import shutil
from tempfile import mkdtemp
import time
import warnings
from subprocess import run
from endplay.parsers.dealer import DealerParser, ParseException, Node
from endplay.dealer.constraint import ConstraintInterpreter
from endplay.dealer.actions import TerminalActions, LaTeXActions, HTMLActions
from endplay.dealer.generate import generate_deals
from endplay.types import Deal
import random

class LaTeXError(RuntimeError):
	"Raised when there was an error running pdflatex"
	pass

def run_script(
	script: Optional[str], 
	show_progress: bool = False,
	produce: int = 40,
	generate: int = 1000000,
	seed: int = None,
	verbose: bool = False,
	swapping: int = 0,
	outformat: str = "plain",
	outfile: Optional[str] = None,
	constraints: list[str] = [],
	actions: list[str] = [],
	predeal: str = "",
	board_numbers: bool = False) -> list[Deal]:
	"""
	Execute a dealer script file

	:param script: The name of the script file to run
	:param show_progress: Display a progress meter while hands are generated
	:param produce: The number of hands to produce
	:param generate: The maximum number of shuffles to perform
	:param seed: The seed for the random number generator
	:param verbose: Print extra debugging info and statistics at completion
	:param swapping: The swapping algorithm to use (0=no swapping, 2=swap EW, 3=all permutations of SEW)
	:param outformat: The format to print output in: 'plain', 'latex' or 'pdf'
	:param outfile: A filename to write the output to, if None then printed to stdout
	:param constraints: A list of extra constraints to apply
	:param actions: A list of extra actions to apply
	:param predeal: A list of players and the suit holdings to deal to them
	:param board_numbers: If True, print board numbers along with the generated deals
	:return: The generated deals in a list
	"""

	start_time = time.time()

	# If we are asked to produce more hands than we generate, we will always fail so let's not
	# waste any time trying
	if produce > generate:
		raise ValueError(f"Asked to produce {produce} hands by generating {generate} hands")

	if seed is None:
		# Generate a seed between 0 and 2**32-1 (required by numpy.random.RandomState)
		# so that if the verbose option is passed we can print out the value of the initial
		# seed at the end of the run
		seed = random.randrange((1 << 32) - 1)

	# Interpret constraints and actions
	parser = DealerParser()
	constraints = [parser.parse_expr(c) for c in constraints]
	actions = [child for a in actions for child in parser.parse_string("action " + a).first_child.children]
	if predeal:
		predeal_node = parser.parse_string("predeal " + predeal)
		deal = predeal_node.first_child.first_child.value
	else:
		deal = Deal()

	# Parse script into document tree
	if script is None:
		doctree = Node("root", Node.ROOT)
	else:
		try:
			with open(script) as f:
				doctree = parser.parse_file(f)
		except FileNotFoundError:
			raise RuntimeError(f"{script}: no such file")
		except OSError as e:
			raise RuntimeError(f"Could not load script: {e}")
		except ParseException as e:
			raise RuntimeError(f"Syntax error: {e}")
		except Exception as e:
			raise RuntimeError(f"Unknown exception occurred: {e}")

	# Initialise remaining variables from script
	interp = ConstraintInterpreter()
	vulnerable = None
	player = None
	try:
		for node in doctree.children:
			if node.value == "generate":
				generate = node.first_child.value
			elif node.value == "produce":
				produce = node.first_child.value
			elif node.value == "vulnerable":
				vulnerable = node.first_child.value
			elif node.value == "dealer":
				player = node.first_child.value
			elif node.value == "predeal":
				deal = node.first_child.value
			elif node.value == "pointcount":
				pointcount = [child.value for child in node.children]
				interp.set_env("hcpscale", pointcount)
			elif node.value == "altcount":
				pointcount = [child.value for child in node.children[1:]]
				interp.set_env(f"pt{node.first_child.value}", pointcount)
			elif node.value == "condition":
				constraints += [node.first_child]
			elif node.value == "action":
				actions += [child for child in node.children]
			elif node.value == "define":
				interp.set_env(node.first_child.value, node.last_child)
			else:
				raise RuntimeError("Unknown dealer input:", node.value)
	except NotImplementedError as e:
		exit(f"One of the features you are trying to use is unimplemented: {e}")
	except Exception as e:
		exit(f"Unknown exception occurred: {e}",)

	# Produce hands
	compiled_constraints = [interp.lambdify(c) for c in constraints]
	deals = []
	generator = generate_deals(
		*compiled_constraints, 
		predeal = deal, 
		swapping = swapping, 
		show_progress = show_progress, 
		produce = produce, 
		seed = seed, 
		max_attempts = generate)
	try:
		while True: deals.append(next(generator))
	except StopIteration as e:
		actual_generated = e.value

	# Try and guess the output format
	if outformat is None:
		if isinstance(outfile, str):
			if outfile.endswith(".html") or outfile.endswith(".htm"):
				outformat = "html"
			elif outfile.endswith(".tex"):
				outformat = "latex"
			elif outfile.endswith(".pdf"):
				outformat = "pdf"
			else:
				outformat = "plain"
		else:
			outformat = "plain"

	# Set up the output engine
	if isinstance(outfile, str):
		outfile_name = outfile
		outfile = open(outfile, "w", encoding="utf-8")
	if outformat == "plain":
		actioner = TerminalActions(deals, outfile, board_numbers)
	elif outformat == "latex":
		actioner = LaTeXActions(deals, outfile, board_numbers)
	elif outformat == "pdf":
		if outfile is None:
			raise RuntimeError("Output file must be specified with pdf file format")
		tmpdir = mkdtemp()
		tmpoutfile = open(tmpdir + "/main.tex", "w", encoding="utf-8")
		actioner = LaTeXActions(deals, tmpoutfile, board_numbers)
	elif outformat == "html":
		actioner = HTMLActions(deals, outfile, board_numbers)
	else:
		raise RuntimeError(f"Unknown file format {outformat} specified")
	actioner.write_preamble()

	# Run actions
	if not actions:
		actioner.printall()
	else:
		for action in actions:
			if action.value == "printall":
				actioner.printall()
			elif action.value == "print":
				actioner.print(*[child.value for child in action.children])
			elif action.value == "printew":
				actioner.printew()
			elif action.value == "printpbn":
				actioner.printpbn()
			elif action.value == "printcompact":
				if len(action.children) == 0:
					expr = None
				else:
					expr = interp.lambdify(action.first_child)
				actioner.printcompact(expr)
			elif action.value == "printoneline":
				if len(action.children) == 0:
					expr = None
				else:
					expr = interp.lambdify(action.first_child)
				actioner.printoneline(expr)
			elif action.value == "printes":
				objs = []
				for child in action.children:
					if child.dtype == Node.VALUE:
						objs.append(child.value)
					else:
						objs.append(interp.lambdify(child))
				actioner.printes(*objs)
			elif action.value == "average":
				if len(action.children) == 2:
					s, expr = action.first_child.value, interp.lambdify(action.last_child)
				else:
					s, expr = None, interp.lambdify(action.last_child)
				actioner.average(expr, s)
			elif action.value == "frequency":
				if action.first_child.dtype == Node.VALUE:
					s, args = action.children[0].value, action.children[1:]
				else:
					s, args = None, action.children
				hargs = [interp.lambdify(args[0]), args[1].value, args[2].value]
				if len(args) > 3:
					hargs += [interp.lambdify(args[3]), args[4].value, args[5].value]
					actioner.frequency2d(*hargs, s)
				else:
					actioner.frequency1d(*hargs, s)
			else:
				raise ValueError(f"Unknown action {action.value}")

	# Close the output engine
	actioner.write_postamble()
	if outfile:
		outfile.close()

	if outformat == "pdf":
		tmpoutfile.close()
		proc = run(["pdflatex", "main.tex"], cwd=tmpdir, input=b'', capture_output=True)
		if proc.returncode != 0:
			with open(tmpdir + "/main.log") as f:
				log = f.read()
			with open(tmpdir + "/main.tex") as f:
				tex = f.read()
			raise LaTeXError("LaTeX error:main.log:\n" + log + "\ninput file:\n" + tex)
		shutil.copy2(tmpdir + "/main.pdf", outfile_name)
		try:
			shutil.rmtree(tmpdir)
		except Exception as e:
			warnings.warn(f"Unable to remove temporary directory tree {tmpdir}: {e}", ResourceWarning)

	if verbose:
		print("Generated", actual_generated, "hands")
		print("Produced", len(deals), "hands")
		print("Initial random seed", seed)
		print(f"Time needed {time.time()-start_time:.3f}s")

	return deals